from __future__ import annotations

import inspect
from typing import Literal

import stim

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.compile.specs.base import CubeSpec
from tqec.compile.specs.enums import SpatialArms
from tqec.compile.specs.library.generators.utils import PlaquetteMapper
from tqec.plaquette.compilation.base import PlaquetteCompiler
from tqec.plaquette.debug import DrawPolygon, PlaquetteDebugInformation
from tqec.plaquette.enums import PlaquetteOrientation
from tqec.plaquette.plaquette import Plaquette, Plaquettes
from tqec.plaquette.qubit import SquarePlaquetteQubits
from tqec.plaquette.rpng.rpng import PauliBasis, RPNGDescription
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.templates.base import RectangularTemplate
from tqec.templates.qubit import (
    QubitHorizontalBorders,
    QubitSpatialCubeTemplate,
    QubitTemplate,
    QubitVerticalBorders,
)
from tqec.utils.enums import Basis, Orientation
from tqec.utils.exceptions import TQECError
from tqec.utils.frozendefaultdict import FrozenDefaultDict


def make_fixed_bulk_realignment_plaquette(
    stabilizer_basis: Basis,
    z_orientation: Orientation,
    mq_reset: Basis,
    mq_measurement: Basis,
    debug_basis: PauliBasis | None = None,
) -> Plaquette:
    """Make the plaquette used for fixed-bulk temporal Hadamard transition."""
    qubits = SquarePlaquetteQubits()
    cx_targets: list[tuple[int, int]]
    # used to match the 5-timestep schedule in the other part of computation
    cx_schedule: list[int]
    match stabilizer_basis, z_orientation:
        case Basis.Z, Orientation.VERTICAL:
            cx_targets = [(0, 4), (1, 4), (4, 2), (4, 0)]
            cx_schedule = [1, 2, 3, 5]
        case Basis.Z, Orientation.HORIZONTAL:
            cx_targets = [(0, 4), (2, 4), (4, 1), (4, 0)]
            cx_schedule = [1, 3, 4, 5]
        case Basis.X, Orientation.VERTICAL:
            cx_targets = [(4, 0), (4, 2), (1, 4), (0, 4)]
            cx_schedule = [1, 3, 4, 5]
        case Basis.X, Orientation.HORIZONTAL:
            cx_targets = [(4, 0), (4, 1), (2, 4), (0, 4)]
            cx_schedule = [1, 2, 3, 5]
    circuit = stim.Circuit()
    circuit.append(f"R{mq_reset.value}", qubits.syndrome_qubits_indices, [])
    circuit.append("TICK", [], [])
    for targets in cx_targets:
        circuit.append("CX", targets, [])
        circuit.append("TICK", [], [])
    circuit.append(f"M{mq_measurement.value}", qubits.syndrome_qubits_indices, [])
    circuit.append("H", qubits.data_qubits_indices, [])
    schedule = [0, *cx_schedule, 6]
    scheduled_circuit = ScheduledCircuit.from_circuit(circuit, schedule, qubits.qubit_map)
    return Plaquette(
        f"fixed_bulk_realignment_{stabilizer_basis}_{z_orientation.value}_R{mq_reset}_M{mq_measurement}",
        qubits,
        scheduled_circuit,
        mergeable_instructions=frozenset({"H"}),
        debug_information=PlaquetteDebugInformation(
            draw_polygons=(DrawPolygon(debug_basis) if debug_basis is not None else None)
        ),
    )


class FixedBulkConventionGenerator:
    def __init__(self, translator: RPNGTranslator, compiler: PlaquetteCompiler):
        """Contain the plaquette generation procedures to implement the fixed bulk convention.

        Args:
            translator: instance used to translate :class:`.RPNGDescription` instances into
                :class:`.Plaquette` instances.
            compiler: instance used to transform :class:`.Plaquette` instances into other
                :class:`.Plaquette` instances.

        """
        self._mapper = PlaquetteMapper(translator, compiler)

    def _not_implemented_exception(self) -> NotImplementedError:
        calling_method_name = inspect.stack(context=0)[1].function
        class_name = type(self).__name__
        return NotImplementedError(
            f"The method '{class_name}.{calling_method_name}' has not been "
            "implemented but is required to continue. Please implement it."
        )

    def get_bulk_rpng_descriptions(
        self,
        reset: Basis | None = None,
        measurement: Basis | None = None,
        reset_and_measured_indices: tuple[Literal[0, 1, 2, 3], ...] = (0, 1, 2, 3),
    ) -> dict[Basis, dict[Orientation, RPNGDescription]]:
        """Get plaquettes that are supposed to be used in the bulk.

        This function returns the four 4-body stabilizer measurement plaquettes
        containing 5 rounds that can be arbitrarily tiled without any gate schedule
        clash. These plaquettes are organised by basis and hook orientation.

        Args:
            reset: basis of the reset operation performed on data-qubits. Defaults
                to ``None`` that translates to no reset being applied on data-qubits.
            measurement: basis of the measurement operation performed on data-qubits.
                Defaults to ``None`` that translates to no measurement being applied
                on data-qubits.
            reset_and_measured_indices: data-qubit indices that should be impacted
                by the provided ``reset`` and ``measurement`` values.

        Returns:
            a mapping with 4 plaquettes: one for each basis (either ``X`` or ``Z``)
            and for each hook orientation (either ``HORIZONTAL`` or ``VERTICAL``).

        """
        # _r/_m: reset/measurement basis applied to each data-qubit in
        # reset_and_measured_indices
        _r = reset.value.lower() if reset is not None else "-"
        _m = measurement.value.lower() if measurement is not None else "-"
        # rs/ms: resets/measurements basis applied for each data-qubit
        rs = [_r if i in reset_and_measured_indices else "-" for i in range(4)]
        ms = [_m if i in reset_and_measured_indices else "-" for i in range(4)]
        # 2-qubit gate schedules
        vsched, hsched = (1, 4, 3, 5), (1, 2, 3, 5)
        return {
            Basis.X: {
                Orientation.VERTICAL: RPNGDescription.from_string(
                    " ".join(f"{r}x{s}{m}" for r, s, m in zip(rs, vsched, ms))
                ),
                Orientation.HORIZONTAL: RPNGDescription.from_string(
                    " ".join(f"{r}x{s}{m}" for r, s, m in zip(rs, hsched, ms))
                ),
            },
            Basis.Z: {
                Orientation.VERTICAL: RPNGDescription.from_string(
                    " ".join(f"{r}z{s}{m}" for r, s, m in zip(rs, vsched, ms))
                ),
                Orientation.HORIZONTAL: RPNGDescription.from_string(
                    " ".join(f"{r}z{s}{m}" for r, s, m in zip(rs, hsched, ms))
                ),
            },
        }

    def get_3_body_rpng_descriptions(
        self,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> tuple[RPNGDescription, RPNGDescription, RPNGDescription, RPNGDescription]:
        """Return the four 3-body stabilizer measurement plaquettes.

        Args:
            reset: basis of the reset operation performed on data-qubits. Defaults
                to ``None`` that translates to no reset being applied on data-qubits.
            measurement: basis of the measurement operation performed on data-qubits.
                Defaults to ``None`` that translates to no measurement being applied
                on data-qubits.

        Returns:
            the four 3-body stabilizer measurement plaquettes. Their order follow the usual
            convention: ``(top_left, top_right, bottom_left, bottom_right)``.

        """
        # r/m: reset/measurement basis applied to each data-qubit
        r = reset.value.lower() if reset is not None else "-"
        m = measurement.value.lower() if measurement is not None else "-"
        # Note: the schedule of CNOT gates in corner plaquettes is less important
        # because hook errors do not exist on 3-body stabilizers. We arbitrarily
        # chose the schedule of the plaquette group the corner belongs to.
        # Note that we include resets and measurements on all the used data-qubits.
        # That should be fine because this plaquette only touches cubes and pipes
        # that are related to the spatial junction being implemented, and it is not
        # valid to have a temporal pipe coming from below a spatial junction, hence
        # the data-qubits cannot be already initialised to a value we would like to
        # keep and that would be destroyed by reset/measurement.
        return (
            RPNGDescription.from_string(f"---- {r}z4{m} {r}z3{m} {r}z5{m}"),
            RPNGDescription.from_string(f"{r}x1{m} ---- {r}x3{m} {r}x5{m}"),
            RPNGDescription.from_string(f"{r}x1{m} {r}x2{m} ---- {r}x5{m}"),
            RPNGDescription.from_string(f"{r}z1{m} {r}z4{m} {r}z3{m} ----"),
        )

    def get_2_body_rpng_descriptions(
        self,
    ) -> dict[Basis, dict[PlaquetteOrientation, RPNGDescription]]:
        """Get plaquettes that are supposed to be used on the boundaries.

        This function returns the eight 2-body stabilizer measurement plaquettes
        that can be used on the 5-round plaquettes returned by
        :meth:`get_bulk_plaquettes`.

        Note:
            The 2-body stabilizer measurement plaquettes returned by this function
            all follow the same schedule: ``1-2-3-5``.

        Warning:
            By convention, the 2-body stabilizers never reset/measure any
            data-qubit. This is done because it is way simpler to reset the correct
            data-qubits in 4-body stabilizers, and the resets/measurements in 2-body
            stabilizers would be redundant.

        Warning:
            This function uses the :class:`~tqec.plaquette.enums.PlaquetteOrientation`
            class. For a 2-body stabilizer measurement plaquette, the "orientation"
            corresponds to the direction in which the rounded side is pointing.
            So a plaquette with the orientation ``PlaquetteOrientation.DOWN`` has the
            following properties:
            - it measures the 2 data-qubits on the **top** side of the usual 4-body
            stabilizer measurement plaquette,
            - it can be used for a bottom boundary,
            - its rounded side points downwards.

        Returns:
            a mapping with 8 plaquettes: one for each basis (either ``X`` or ``Z``)
            and for each plaquette orientation (``UP``, ``DOWN``, ``LEFT`` or
            ``RIGHT``).

        """
        return {
            Basis.X: {
                PlaquetteOrientation.DOWN: RPNGDescription.from_string("-x1- -x2- ---- ----"),
                PlaquetteOrientation.LEFT: RPNGDescription.from_string("---- -x2- ---- -x5-"),
                PlaquetteOrientation.UP: RPNGDescription.from_string("---- ---- -x3- -x5-"),
                PlaquetteOrientation.RIGHT: RPNGDescription.from_string("-x1- ---- -x3- ----"),
            },
            Basis.Z: {
                PlaquetteOrientation.DOWN: RPNGDescription.from_string("-z1- -z2- ---- ----"),
                PlaquetteOrientation.LEFT: RPNGDescription.from_string("---- -z2- ---- -z5-"),
                PlaquetteOrientation.UP: RPNGDescription.from_string("---- ---- -z3- -z5-"),
                PlaquetteOrientation.RIGHT: RPNGDescription.from_string("-z1- ---- -z3- ----"),
            },
        }

    ############################################################
    #                          Memory                          #
    ############################################################

    ########################################
    #             Regular qubit            #
    ########################################
    def get_memory_qubit_raw_template(self) -> RectangularTemplate:
        """Return the template instance needed to implement a single logical qubit."""
        return QubitTemplate()

    def get_memory_qubit_rpng_descriptions(
        self,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a standard memory operation
        on a logical qubit.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_memory_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned logical qubit description.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard
            memory operation on a logical qubit, optionally with resets or
            measurements on the data-qubits too.

        """
        # Border plaquette indices
        up, down, left, right = (
            (6, 13, 7, 12) if z_orientation == Orientation.VERTICAL else (5, 14, 8, 11)
        )
        # Basis for top/bottom and left/right boundary plaquettes
        hbasis = Basis.Z if z_orientation == Orientation.HORIZONTAL else Basis.X
        vbasis = hbasis.flipped()
        # Hook errors orientations
        zhook = z_orientation.flip()
        xhook = zhook.flip()
        # Get plaquette descriptions we will need later
        bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement)
        two_body_descriptions = self.get_2_body_rpng_descriptions()
        return FrozenDefaultDict(
            {
                up: two_body_descriptions[vbasis][PlaquetteOrientation.UP],
                left: two_body_descriptions[hbasis][PlaquetteOrientation.LEFT],
                # Bulk
                9: bulk_descriptions[Basis.Z][zhook],
                10: bulk_descriptions[Basis.X][xhook],
                right: two_body_descriptions[hbasis][PlaquetteOrientation.RIGHT],
                down: two_body_descriptions[vbasis][PlaquetteOrientation.DOWN],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_memory_qubit_plaquettes(
        self,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a standard memory operation on a logical
        qubit.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_memory_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned logical qubit description.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard
            memory operation on a logical qubit, optionally with resets or
            measurements on the data-qubits too.

        """
        return self._mapper(self.get_memory_qubit_rpng_descriptions)(
            z_orientation, reset, measurement
        )

    ########################################
    #                X pipe                #
    ########################################
    def get_memory_vertical_boundary_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a regular spatial pipe between two logical qubits aligned on the ``X`` axis.
        """
        return QubitVerticalBorders()

    def get_memory_vertical_boundary_rpng_descriptions(
        self,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a standard memory operation
        on a pipe between two neighbouring logical qubits aligned on the ``X``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_memory_vertical_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard memory
            operation on a pipe between two neighbouring logical qubits aligned on
            the ``X``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        # Border plaquette indices
        up, down = (2, 3) if z_orientation == Orientation.VERTICAL else (1, 4)
        # Basis for top/bottom boundary plaquettes
        vbasis = Basis.Z if z_orientation == Orientation.VERTICAL else Basis.X
        # Hook errors orientations
        zhook = z_orientation.flip()
        xhook = zhook.flip()
        # Generating the plaquette descriptions we will need later
        left_bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (1, 3))
        right_bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (0, 2))
        two_body_descriptions = self.get_2_body_rpng_descriptions()

        return FrozenDefaultDict(
            {
                up: two_body_descriptions[vbasis][PlaquetteOrientation.UP],
                down: two_body_descriptions[vbasis][PlaquetteOrientation.DOWN],
                # LEFT bulk
                5: left_bulk_descriptions[Basis.Z][zhook],
                6: left_bulk_descriptions[Basis.X][xhook],
                # RIGHT bulk
                7: right_bulk_descriptions[Basis.X][xhook],
                8: right_bulk_descriptions[Basis.Z][zhook],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_memory_vertical_boundary_plaquettes(
        self,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a standard memory operation on a pipe between
        two neighbouring logical qubits aligned on the ``X``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_memory_vertical_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a standard memory operation on a
            pipe between two neighbouring logical qubits aligned on the
            ``X``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        return self._mapper(self.get_memory_vertical_boundary_rpng_descriptions)(
            z_orientation, reset, measurement
        )

    ########################################
    #                Y pipe                #
    ########################################
    def get_memory_horizontal_boundary_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a regular spatial pipe between two logical qubits aligned on the ``Y`` axis.
        """
        return QubitHorizontalBorders()

    def get_memory_horizontal_boundary_rpng_descriptions(
        self,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a standard memory operation
        on a pipe between two neighbouring logical qubits aligned on the ``Y``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_memory_horizontal_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard memory
            operation on a pipe between two neighbouring logical qubits aligned on
            the ``Y``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        # Border plaquette indices
        left, right = (1, 4) if z_orientation == Orientation.VERTICAL else (3, 2)
        # Basis for left/right boundary plaquettes
        hbasis = Basis.Z if z_orientation == Orientation.HORIZONTAL else Basis.X
        # Hook errors orientations
        zhook = z_orientation.flip()
        xhook = zhook.flip()
        # Generating the plaquette descriptions we will need later
        up_bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (2, 3))
        down_bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (0, 1))
        two_body_descriptions = self.get_2_body_rpng_descriptions()

        return FrozenDefaultDict(
            {
                left: two_body_descriptions[hbasis][PlaquetteOrientation.LEFT],
                right: two_body_descriptions[hbasis][PlaquetteOrientation.RIGHT],
                # TOP bulk
                5: up_bulk_descriptions[Basis.Z][zhook],
                6: up_bulk_descriptions[Basis.X][xhook],
                # BOTTOM bulk
                7: down_bulk_descriptions[Basis.X][xhook],
                8: down_bulk_descriptions[Basis.Z][zhook],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_memory_horizontal_boundary_plaquettes(
        self,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a standard memory operation on a pipe between
        two neighbouring logical qubits aligned on the ``Y``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_memory_horizontal_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a standard memory operation on a
            pipe between two neighbouring logical qubits aligned on the
            ``Y``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        return self._mapper(self.get_memory_horizontal_boundary_rpng_descriptions)(
            z_orientation, reset, measurement
        )

    ############################################################
    #                          Spatial                         #
    ############################################################
    def get_spatial_cube_qubit_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC gadgets
            such as the lattice surgery implementation of a ``CZ`` gate.

        """
        return QubitSpatialCubeTemplate()

    def get_spatial_cube_qubit_rpng_descriptions(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC
            gadgets such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_spatial_cube_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: flag-like enumeration listing the arms that are used around
                the logical qubit. The returned template will be adapted to be
                compatible with such a layout.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arms`` only contains 0 or 1 flag.
            TQECError: if ``arms`` describes an I-shaped junction (TOP/DOWN
                or LEFT/RIGHT).

        Returns:
            a description of the plaquettes needed to implement a spatial cube.

        """
        # In this function implementation, all the indices used are referring to the
        # indices returned by the QubitSpatialCubeTemplate template. They are
        # copied below for convenience, but the only source of truth is in the
        # QubitSpatialCubeTemplate docstring!
        #      1   9  10   9  10   9  10   9  10   2
        #     11   5  17  13  17  13  17  13   6  21
        #     12  20  13  17  13  17  13  17  14  22
        #     11  16  20  13  17  13  17  14  18  21
        #     12  20  16  20  13  17  14  18  14  22
        #     11  16  20  16  19  15  18  14  18  21
        #     12  20  16  19  15  19  15  18  14  22
        #     11  16  19  15  19  15  19  15  18  21
        #     12   7  15  19  15  19  15  19   8  22
        #      3  23  24  23  24  23  24  23  24   4
        if arms in SpatialArms.I_shaped_arms():
            raise TQECError(
                "I-shaped spatial junctions (i.e., spatial junctions with only two "
                "arms that are the opposite of each other: LEFT/RIGHT or UP/DOWN) "
                "should not use get_spatial_cube_qubit_template but rather use "
                "a conventional memory logical qubit with get_memory_qubit_template."
            )
        # Get parity information in a more convenient format.
        boundary_is_z = spatial_boundary_basis == Basis.Z
        # Pre-define some collection of plaquette descriptions
        corner_descriptions = self.get_3_body_rpng_descriptions(reset, measurement)
        bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement)
        two_body_descriptions = self.get_2_body_rpng_descriptions()

        mapping: dict[int, RPNGDescription] = {}

        ####################
        #    Boundaries    #
        ####################
        # Fill the boundaries that should be filled in the returned template
        # because they have no arms, and so will not be filled later.
        # Note that indices 1, 2, 3 and 4 **might** be set twice in the 4 ifs
        # below. These cases are handled later in the function and will
        # remove the description on 1, 2, 3 or 4 if needed, so we do not have
        # to account for those cases here.
        # Note that resets and measurements are included on all data-qubits here.
        # Alias for readability: _sbb <=> Spatial Boundary Basis.
        _sbb = spatial_boundary_basis
        if SpatialArms.UP not in arms:
            corner, bulk = (1, 10) if boundary_is_z else (2, 9)
            mapping[corner] = mapping[bulk] = two_body_descriptions[_sbb][PlaquetteOrientation.UP]
        if SpatialArms.RIGHT not in arms:
            corner, bulk = (4, 21) if boundary_is_z else (2, 22)
            mapping[corner] = mapping[bulk] = two_body_descriptions[_sbb][
                PlaquetteOrientation.RIGHT
            ]
        if SpatialArms.DOWN not in arms:
            corner, bulk = (4, 23) if boundary_is_z else (3, 24)
            mapping[corner] = mapping[bulk] = two_body_descriptions[_sbb][PlaquetteOrientation.DOWN]
        if SpatialArms.LEFT not in arms:
            corner, bulk = (1, 12) if boundary_is_z else (3, 11)
            mapping[corner] = mapping[bulk] = two_body_descriptions[_sbb][PlaquetteOrientation.LEFT]

        # For each corner, if the two arms around the corner are not present, the
        # corner plaquette should be removed from the mapping (this is the case
        # where it has been set twice in the ifs above).
        if SpatialArms.LEFT not in arms and SpatialArms.UP not in arms and boundary_is_z:
            del mapping[1]
        if SpatialArms.UP not in arms and SpatialArms.RIGHT not in arms and not boundary_is_z:
            del mapping[2]
        if SpatialArms.DOWN not in arms and SpatialArms.LEFT not in arms and not boundary_is_z:
            del mapping[3]
        if SpatialArms.RIGHT not in arms and SpatialArms.DOWN not in arms and boundary_is_z:
            del mapping[4]

        ####################
        #       Bulk       #
        ####################
        # Assigning plaquette description to the bulk, considering that the bulk
        # corners (i.e. indices {5, 6, 7, 8}) should be assigned "regular" plaquettes
        # (i.e. 6 is assigned the same plaquette as 17, 7 -> 19, 5 -> 13, 8 -> 15).
        # If these need to be changed, it will be done afterwards.
        # Setting the orientations for Z plaquettes for each of the four portions of
        # the template bulk.
        zup = zdown = Orientation.VERTICAL if boundary_is_z else Orientation.HORIZONTAL
        zright = zleft = zup.flip()
        # If the corresponding arm is missing, the Z plaquette hook error orientation
        # should flip to avoid shortcuts due to hook errors.
        zup = zup if SpatialArms.UP in arms else zup.flip()
        zdown = zdown if SpatialArms.DOWN in arms else zdown.flip()
        zright = zright if SpatialArms.RIGHT in arms else zright.flip()
        zleft = zleft if SpatialArms.LEFT in arms else zleft.flip()
        # The X orientations are the opposite of the Z orientation
        xup, xdown, xright, xleft = (zup.flip(), zdown.flip(), zright.flip(), zleft.flip())

        # Setting the Z plaquettes
        mapping[5] = mapping[13] = bulk_descriptions[Basis.Z][zup]
        mapping[8] = mapping[15] = bulk_descriptions[Basis.Z][zdown]
        mapping[14] = bulk_descriptions[Basis.Z][zright]
        mapping[16] = bulk_descriptions[Basis.Z][zleft]
        # Setting the X plaquettes
        mapping[6] = mapping[17] = bulk_descriptions[Basis.X][xup]
        mapping[7] = mapping[19] = bulk_descriptions[Basis.X][xdown]
        mapping[18] = bulk_descriptions[Basis.X][xright]
        mapping[20] = bulk_descriptions[Basis.X][xleft]

        # For each corner, if the two arms around the corner are not present, the
        # corner plaquette has been removed from the mapping. The corner **within
        # the bulk** should be overwritten to become a 3-body stabilizer measurement.
        # Note that this is not done when deleting the external corners before because
        # the bulk plaquettes are set just above, and so we should override the
        # plaquettes after.
        if SpatialArms.LEFT not in arms and SpatialArms.UP not in arms and boundary_is_z:
            mapping[5] = corner_descriptions[0]
        if SpatialArms.UP not in arms and SpatialArms.RIGHT not in arms and not boundary_is_z:
            mapping[6] = corner_descriptions[1]
        if SpatialArms.DOWN not in arms and SpatialArms.LEFT not in arms and not boundary_is_z:
            mapping[7] = corner_descriptions[2]
        if SpatialArms.RIGHT not in arms and SpatialArms.DOWN not in arms and boundary_is_z:
            mapping[8] = corner_descriptions[3]

        ####################
        #  Sanity checks   #
        ####################
        # All the plaquettes in the bulk should be set.
        bulk_plaquette_indices = set(range(5, 9)) | set(range(13, 21))
        missing_bulk_plaquette_indices = bulk_plaquette_indices - mapping.keys()
        assert not missing_bulk_plaquette_indices, (
            "Some plaquette(s) in the bulk were not correctly assigned to a "
            f"RPNGDescription. Missing indices: {missing_bulk_plaquette_indices}."
        )
        return FrozenDefaultDict(mapping, default_value=RPNGDescription.empty())

    def get_spatial_cube_qubit_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC
            gadgets such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_spatial_cube_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: flag-like enumeration listing the arms that are used around
                the logical qubit. The returned template will be adapted to be
                compatible with such a layout.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arms`` only contains 0 or 1 flag.
            TQECError: if ``arms`` describes an I-shaped junction (TOP/DOWN
                or LEFT/RIGHT).

        Returns:
            the plaquettes needed to implement a spatial cube.

        """
        return self._mapper(self.get_spatial_cube_qubit_rpng_descriptions)(
            spatial_boundary_basis, arms, reset, measurement
        )

    ########################################
    #              Spatial arm             #
    ########################################
    def get_spatial_cube_arm_raw_template(self, arms: SpatialArms) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement the given spatial ``arms``.

        Args:
            arms: specification of the spatial arm(s) we want a template for.
                Needs to contain either one arm, or 2 arms that form a line
                (e.g., ``SpatialArms.UP | SpatialArms.DOWN``).

        """
        if (
            len(arms) == 0
            or len(arms) > 2
            or (len(arms) == 2 and arms not in SpatialArms.I_shaped_arms())
        ):
            raise TQECError(
                f"The two provided arms cannot form a spatial pipe. Got {arms} but "
                f"expected either a single {SpatialArms.__name__} or two but in a "
                f"line (e.g., {SpatialArms.I_shaped_arms()})."
            )
        if SpatialArms.LEFT in arms or SpatialArms.RIGHT in arms:
            return QubitVerticalBorders()
        elif SpatialArms.UP is arms or SpatialArms.DOWN in arms:
            return QubitHorizontalBorders()
        else:
            raise TQECError(f"Unrecognized spatial arm(s): {arms}.")

    def get_spatial_cube_arm_rpng_descriptions(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement **one** pipe connecting to a
        spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC gadgets
            such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_spatial_cube_arm_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: arm(s) of the spatial cube(s) linked by the pipe.
            linked_cubes: a tuple ``(u, v)`` where ``u`` and ``v`` are the
                specifications of the two ends of the pipe to generate RPNG
                descriptions for.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arm`` does not contain exactly 1 or 2 flags (i.e.,
                if it contains 0 or 3+ flags).

        Returns:
            a description of the plaquettes needed to implement **one** pipe
            connecting to a spatial cube.

        """
        if len(arms) == 2 and arms not in SpatialArms.I_shaped_arms():
            raise TQECError(
                f"The two provided arms cannot form a spatial pipe. Got {arms} but "
                f"expected either a single {SpatialArms.__name__} or two but in a "
                f"line (e.g., {SpatialArms.I_shaped_arms()})."
            )
        if arms in [
            SpatialArms.LEFT,
            SpatialArms.RIGHT,
            SpatialArms.LEFT | SpatialArms.RIGHT,
        ]:
            return self._get_left_right_spatial_cube_arm_rpng_descriptions(
                spatial_boundary_basis, arms, linked_cubes, reset, measurement
            )
        if arms in [
            SpatialArms.UP,
            SpatialArms.DOWN,
            SpatialArms.UP | SpatialArms.DOWN,
        ]:
            return self._get_up_down_spatial_cube_arm_rpng_descriptions(
                spatial_boundary_basis, arms, linked_cubes, reset, measurement
            )
        raise TQECError(f"Got an invalid arm: {arms}.")

    def get_spatial_cube_arm_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement **one** pipe connecting to a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC gadgets
            such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBulkConventionGenerator.get_spatial_cube_arm_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: arm(s) of the spatial cube(s) linked by the pipe.
            linked_cubes: a tuple ``(u, v)`` where ``u`` and ``v`` are the
                specifications of the two ends of the pipe to generate RPNG
                descriptions for.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arm`` does not contain exactly 1 or 2 flags (i.e.,
                if it contains 0 or 3+ flags).

        Returns:
            the plaquettes needed to implement **one** pipe connecting to a
            spatial cube.

        """
        return self._mapper(self.get_spatial_cube_arm_rpng_descriptions)(
            spatial_boundary_basis, arms, linked_cubes, reset, measurement
        )

    def _get_left_right_spatial_cube_arm_rpng_descriptions(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        # We need the bulk plaquettes to only reset the central qubits of the pipe.
        # To do so, we have two sets of bulk plaquettes with different reset/measured
        # qubits. Plaquettes that should go on the LEFT part of the pipe should
        # measure right qubits (i.e., indices 1 and 3) and conversely for the RIGHT
        # part.
        # Generating plaquette descriptions we will need later.
        left_boundary_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (1, 3))
        right_boundary_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (0, 2))
        two_body_descriptions = self.get_2_body_rpng_descriptions()
        # The hook errors also need to be adapted to the boundary basis.
        zhook = (
            Orientation.HORIZONTAL if spatial_boundary_basis == Basis.Z else Orientation.VERTICAL
        )
        xhook = zhook.flip()
        # List the plaquettes used. This mapping might be corrected afterwards to
        # avoid overwriting 3-body stabilizers introduced by the spatial cube.
        up = 2 if spatial_boundary_basis == Basis.Z else 1
        down = 3 if spatial_boundary_basis == Basis.Z else 4

        mapping = {
            # Boundaries
            # For convenience of definition here, the 2-body stabilizers never reset
            # or measure their data-qubits. If a reset/measurement is needed, it is
            # already included in the plaquettes in the bulk.
            up: two_body_descriptions[spatial_boundary_basis][PlaquetteOrientation.UP],
            down: two_body_descriptions[spatial_boundary_basis][PlaquetteOrientation.DOWN],
            # LEFT bulk
            5: left_boundary_descriptions[Basis.Z][zhook],
            6: left_boundary_descriptions[Basis.X][xhook],
            # RIGHT bulk
            7: right_boundary_descriptions[Basis.X][xhook],
            8: right_boundary_descriptions[Basis.Z][zhook],
        }

        _corners = self.get_3_body_rpng_descriptions()
        u, v = linked_cubes
        # Alias to reduce clutter in the implementation for corners
        _sbb = spatial_boundary_basis
        # Replaces the top plaquette if it should be a 3-body stabilizer.
        if SpatialArms.LEFT in arms and _sbb == Basis.Z and SpatialArms.UP in v.spatial_arms:
            mapping[up] = _corners[0]
        if SpatialArms.RIGHT in arms and _sbb == Basis.X and SpatialArms.UP in u.spatial_arms:
            mapping[up] = _corners[1]
        # Replaces the bottom plaquette if it should be a 3-body stabilizer.
        if SpatialArms.LEFT in arms and _sbb == Basis.X and SpatialArms.DOWN in v.spatial_arms:
            mapping[down] = _corners[2]
        if SpatialArms.RIGHT in arms and _sbb == Basis.Z and SpatialArms.DOWN in u.spatial_arms:
            mapping[down] = _corners[3]

        return FrozenDefaultDict(mapping, default_value=RPNGDescription.empty())

    def _get_up_down_spatial_cube_arm_rpng_descriptions(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        # We need the bulk plaquettes to only reset the central qubits of the pipe.
        # To do so, we have two sets of bulk plaquettes with different reset/measured
        # qubits. Plaquettes that should go on the UP part of the pipe should measure
        # bottom qubits (i.e., indices 2 and 3) and conversely for the DOWN part.
        # Generating plaquette descriptions we will need later.
        up_bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (2, 3))
        down_bulk_descriptions = self.get_bulk_rpng_descriptions(reset, measurement, (0, 1))
        two_body_description = self.get_2_body_rpng_descriptions()
        # The hook errors also need to be adapted to the boundary basis.
        zhook = (
            Orientation.VERTICAL if spatial_boundary_basis == Basis.Z else Orientation.HORIZONTAL
        )
        xhook = zhook.flip()
        # List the plaquettes used. This mapping might be corrected afterwards to
        # avoid overwriting 3-body stabilizers introduced by the spatial cube.
        left = 3 if spatial_boundary_basis == Basis.Z else 1
        right = 2 if spatial_boundary_basis == Basis.Z else 4

        mapping = {
            # Boundaries
            # For convenience of definition here, the 2-body stabilizers never reset
            # or measure their data-qubits. If a reset/measurement is needed, it is
            # already included in the plaquettes in the bulk.
            left: two_body_description[spatial_boundary_basis][PlaquetteOrientation.LEFT],
            right: two_body_description[spatial_boundary_basis][PlaquetteOrientation.RIGHT],
            # TOP bulk
            5: up_bulk_descriptions[Basis.Z][zhook],
            6: up_bulk_descriptions[Basis.X][xhook],
            # BOTTOM bulk
            7: down_bulk_descriptions[Basis.X][xhook],
            8: down_bulk_descriptions[Basis.Z][zhook],
        }

        corners = self.get_3_body_rpng_descriptions()
        u, v = linked_cubes
        # Aliases to reduce clutter in the implementation for corners
        _sbb = spatial_boundary_basis
        # Replaces the top plaquette if it should be a 3-body stabilizer.
        if SpatialArms.UP in arms and _sbb == Basis.Z and SpatialArms.LEFT in v.spatial_arms:
            mapping[left] = corners[0]
        if SpatialArms.UP in arms and _sbb == Basis.X and SpatialArms.RIGHT in v.spatial_arms:
            mapping[right] = corners[1]
        # Replaces the bottom plaquette if it should be a 3-body stabilizer.
        if SpatialArms.DOWN in arms and _sbb == Basis.X and SpatialArms.LEFT in u.spatial_arms:
            mapping[left] = corners[2]
        if SpatialArms.DOWN in arms and _sbb == Basis.Z and SpatialArms.RIGHT in u.spatial_arms:
            mapping[right] = corners[3]

        return FrozenDefaultDict(mapping, default_value=RPNGDescription.empty())

    ############################################################
    #                         Hadamard                         #
    ############################################################

    def get_temporal_hadamard_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.Template` instance needed to implement a
        transversal Hadamard gate applied on one logical qubit.
        """
        return QubitTemplate()

    def get_temporal_hadamard_realignment_plaquettes(
        self, z_orientation: Orientation = Orientation.HORIZONTAL
    ) -> Plaquettes:
        """Return the :class:`.Plaquettes` instance needed to implement the realignment of the bulk
        stabilizer basis of the code.

        This is needed because a transversal Hadamard layer will change the bulk stabilizer basis of
        the code. Under fixed-bulk convention, we use an extra realignment layer to realign the bulk
        stabilizer basis of the code to the original one.

        """
        # plaquettes at the bulk
        x_bulk = make_fixed_bulk_realignment_plaquette(
            stabilizer_basis=Basis.X,
            z_orientation=z_orientation,
            mq_reset=Basis.X,
            mq_measurement=Basis.Z,
            debug_basis=PauliBasis.X,
        )
        z_bulk = make_fixed_bulk_realignment_plaquette(
            stabilizer_basis=Basis.Z,
            z_orientation=z_orientation,
            mq_reset=Basis.Z,
            mq_measurement=Basis.X,
            debug_basis=PauliBasis.Z,
        )
        # plaquettes at the right boundary of the template
        right_boundary_basis = Basis.Z if z_orientation == Orientation.HORIZONTAL else Basis.X
        x_right = make_fixed_bulk_realignment_plaquette(
            stabilizer_basis=Basis.X,
            z_orientation=z_orientation,
            mq_reset=right_boundary_basis,
            mq_measurement=right_boundary_basis,
            debug_basis=PauliBasis.X if z_orientation == Orientation.VERTICAL else None,
        ).project_on_boundary(PlaquetteOrientation.RIGHT)
        z_right = make_fixed_bulk_realignment_plaquette(
            stabilizer_basis=Basis.Z,
            z_orientation=z_orientation,
            mq_reset=right_boundary_basis,
            mq_measurement=right_boundary_basis,
            debug_basis=PauliBasis.Z if z_orientation == Orientation.HORIZONTAL else None,
        ).project_on_boundary(PlaquetteOrientation.RIGHT)
        down_boundary_basis = right_boundary_basis.flipped()
        x_down = make_fixed_bulk_realignment_plaquette(
            stabilizer_basis=Basis.X,
            z_orientation=z_orientation,
            mq_reset=down_boundary_basis,
            mq_measurement=down_boundary_basis,
            debug_basis=PauliBasis.X if z_orientation == Orientation.HORIZONTAL else None,
        ).project_on_boundary(PlaquetteOrientation.DOWN)
        z_down = make_fixed_bulk_realignment_plaquette(
            stabilizer_basis=Basis.Z,
            z_orientation=z_orientation,
            mq_reset=down_boundary_basis,
            mq_measurement=down_boundary_basis,
            debug_basis=PauliBasis.Z if z_orientation == Orientation.VERTICAL else None,
        ).project_on_boundary(PlaquetteOrientation.DOWN)
        realign_mapping = FrozenDefaultDict(
            {9: z_bulk, 10: x_bulk, 11: z_right, 12: x_right, 13: z_down, 14: x_down},
            default_value=self._mapper.get_plaquette(RPNGDescription.empty()),
        )
        return Plaquettes(realign_mapping)

    ########################################
    #                X pipe                #
    ########################################
    def get_spatial_vertical_hadamard_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.Template` instance needed to implement a spatial
        Hadamard pipe between two logical qubits aligned on the ``X`` axis.
        """
        raise self._not_implemented_exception()

    def get_spatial_vertical_hadamard_rpng_descriptions(
        self,
        top_left_is_z_stabilizer: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a Hadamard spatial transition
        between two neighbouring logical qubits aligned on the ``X`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``X``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`PlaquetteGenerator.get_spatial_vertical_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.Template` instance returned by this
            method.

        Arguments:
            top_left_is_z_stabilizer: if ``True``, the plaquette with index 5 in
                :class:`~tqec.templates.qubit.QubitVerticalBorders`
                should be measuring a ``Z`` stabilizer on its 2 left-most
                data-qubits and a ``X`` stabilizer on its 2 right-most
                data-qubits. Else, it measures a ``X`` stabilizer on its two
                left-most data-qubits and a ``Z`` stabilizer on its two
                right-most data-qubits.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a Hadamard
            spatial transition between two neighbouring logical qubits aligned
            on the ``X`` axis.

        """
        raise self._not_implemented_exception()  # pragma: no cover

    def get_spatial_vertical_hadamard_plaquettes(
        self,
        top_left_is_z_stabilizer: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a Hadamard spatial transition between two
        neighbouring logical qubits aligned on the ``X`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``X``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`PlaquetteGenerator.get_spatial_vertical_hadamard_raw_template` and the returned
            plaquettes should only be considered valid when used in conjunction with the
            :class:`~tqec.templates.base.Template` instance returned by this method.

        Arguments:
            top_left_is_z_stabilizer: if ``True``, the plaquette with index 5 in
                :class:`~tqec.templates.qubit.QubitVerticalBorders`
                should be measuring a ``Z`` stabilizer on its 2 left-most
                data-qubits and a ``X`` stabilizer on its 2 right-most
                data-qubits. Else, it measures a ``X`` stabilizer on its two
                left-most data-qubits and a ``Z`` stabilizer on its two
                right-most data-qubits.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a Hadamard spatial transition between two
            neighbouring logical qubits aligned on the ``X`` axis.

        """
        return self._mapper(self.get_spatial_vertical_hadamard_rpng_descriptions)(
            top_left_is_z_stabilizer, reset, measurement
        )

    ########################################
    #                Y pipe                #
    ########################################
    def get_spatial_horizontal_hadamard_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.Template` instance needed to implement a spatial
        Hadamard pipe between two neighbouring logical qubits aligned on the ``Y`` axis.
        """
        raise self._not_implemented_exception()

    def get_spatial_horizontal_hadamard_rpng_descriptions(
        self,
        top_left_is_z_stabilizer: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a Hadamard spatial transition
        between two neighbouring logical qubits aligned on the ``Y`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``Y``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`PlaquetteGenerator.get_spatial_horizontal_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.Template` instance returned by this
            method.

        Arguments:
            top_left_is_z_stabilizer: if ``True``, the plaquette with index 5 in
                :class:`~tqec.templates.qubit.QubitHorizontalBorders` should be
                measuring a ``Z`` stabilizer on its 2 top-most data-qubits and a
                ``X`` stabilizer on its 2 bottom-most data-qubits. Else, it
                measures a ``X`` stabilizer on its two top-most data-qubits and
                a ``Z`` stabilizer on its two bottom-most data-qubits.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a Hadamard
            spatial transition between two neighbouring logical qubits aligned
            on the ``Y`` axis.

        """
        raise self._not_implemented_exception()  # pragma: no cover

    def get_spatial_horizontal_hadamard_plaquettes(
        self,
        top_left_is_z_stabilizer: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a Hadamard spatial transition between two
        neighbouring logical qubits aligned on the ``Y`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``Y``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`PlaquetteGenerator.get_spatial_horizontal_hadamard_raw_template` and the returned
            plaquettes should only be considered valid when used in conjunction with the
            :class:`~tqec.templates.base.Template` instance returned by this method.

        Arguments:
            top_left_is_z_stabilizer: if ``True``, the plaquette with index 5 in
                :class:`~tqec.templates.qubit.QubitHorizontalBorders` should be
                measuring a ``Z`` stabilizer on its 2 top-most data-qubits and a
                ``X`` stabilizer on its 2 bottom-most data-qubits. Else, it
                measures a ``X`` stabilizer on its two top-most data-qubits and
                a ``Z`` stabilizer on its two bottom-most data-qubits.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a Hadamard spatial transition between two
            neighbouring logical qubits aligned on the ``Y`` axis.

        """
        return self._mapper(self.get_spatial_horizontal_hadamard_rpng_descriptions)(
            top_left_is_z_stabilizer, reset, measurement
        )



"""Defines :class:`.ScheduledCircuit` that represents a quantum circuit with a specific schedule.

This module defines the class used to represent a quantum circuit. It is a
"scheduled" circuit. Such circuits are composed of a finite number of ordered
:class:`~tqec.circuit.moment.Moment` instances that are each associated to an
integer in a strictly increasing list of integers (called "schedule" and
represented by instances of :class:`~.schedule.schedule.Schedule`).

"""

from __future__ import annotations

import bisect
from collections.abc import Callable, Iterable, Iterator, Sequence
from copy import copy, deepcopy
from typing import Any

import stim

from tqec.circuit.moment import Moment, iter_stim_circuit_without_repeat_by_moments
from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap, get_qubit_map
from tqec.circuit.schedule.exception import ScheduleError
from tqec.circuit.schedule.schedule import Schedule
from tqec.utils.exceptions import TQECError
from tqec.utils.instructions import is_annotation_instruction


class ScheduledCircuit:
    def __init__(
        self,
        moments: list[Moment],
        schedule: Schedule | list[int] | int,
        qubit_map: QubitMap,
        _avoid_checks: bool = False,
    ) -> None:
        """Represent a quantum circuit with scheduled moments.

        This class aims at representing a ``stim.Circuit`` instance that has all
        its moments scheduled, i.e., associated with a time slice.

        This class explicitly does not support ``stim.CircuitRepeatBlock`` (i.e.,
        ``REPEAT`` instruction). It will raise an exception if such an
        instruction is found.

        Args:
            moments: moments representing the computation that is scheduled.
                The provided ``moments`` should not contain any ``QUBIT_COORDS``
                instructions. Instead, the qubit coordinates should be provided
                through the ``qubit_map`` parameter.
            schedule: schedule of the provided ``moments``. If an integer is
                provided, each moment of the provided ``stim.Circuit`` is
                scheduled sequentially, starting by the provided integer.
            qubit_map: a map from indices to qubits that is used to re-create
                ``QUBIT_COORDS`` instructions when generating a ``stim.Circuit``.
            _avoid_checks: if True, the inputs are not checked for pre-condition
                violation and it is up to the user to ensure that
                :class:`ScheduledCircuit` pre-conditions are checked by the
                provided input.

        Raises:
            ScheduleError: if the provided ``schedule`` is invalid.
            TQECError: if the provided ``circuit`` contains at least one
                ``stim.CircuitRepeatBlock`` instance.
            TQECError: if the provided ``circuit`` contains at least one
                ``QUBIT_COORDS`` instruction after the first ``TICK`` instruction.

        """
        # Handle each input type explicitly due to type checker
        if isinstance(schedule, int):
            schedule_temp = list(range(schedule, schedule + len(moments)))
            schedule = Schedule(schedule=schedule_temp)
        elif isinstance(schedule, list) and not isinstance(schedule, Schedule):
            schedule = Schedule(schedule=schedule)

        if len(moments) != len(schedule):
            raise ScheduleError(
                "ScheduledCircuit expects all the provided moments to be scheduled. "
                f"Got {len(moments)} moments but {len(schedule)} schedules."
            )
        if not _avoid_checks and any(m.contains_instruction("QUBIT_COORDS") for m in moments):
            raise ScheduleError(
                "ScheduledCircuit instance expects the input `stim.Circuit` to "
                "not contain any QUBIT_COORDS instruction. Found at least one "
                "moment with a QUBIT_COORDS instruction."
            )

        non_empty_moments_indices = [i for i, m in enumerate(moments) if not m.is_empty]
        self._moments: list[Moment] = [moments[i] for i in non_empty_moments_indices]
        self._qubit_map: QubitMap = qubit_map
        self._schedule = Schedule([schedule[i] for i in non_empty_moments_indices])

    @staticmethod
    def empty() -> ScheduledCircuit:
        """Return an empty :class:`ScheduledCircuit` instance."""
        return ScheduledCircuit([], Schedule(), QubitMap(), _avoid_checks=True)

    @staticmethod
    def from_circuit(
        circuit: stim.Circuit,
        schedule: Schedule | list[int] | int = 0,
        qubit_map: QubitMap | None = None,
    ) -> ScheduledCircuit:
        """Build a :class:`ScheduledCircuit` instance from a circuit and a schedule.

        Args:
            circuit: the instance of ``stim.Circuit`` that is scheduled. Should
                not contain any instance of ``stim.CircuitRepeatBlock``.
            schedule: a sorted list of positive integers or a positive integer.
                If a list is given, it should contain as many values as there
                are moments in the provided ``stim.Circuit`` instance. If an
                integer is provided, each moment of the provided ``stim.Circuit``
                is scheduled sequentially, starting by the provided schedule.
            qubit_map: a map from indices to qubits. If None, this function will
                try to extract the qubit coordinates from the ``QUBIT_COORDS``
                instructions found in the provided ``circuit``. Else, the
                provided map will be used.

        Raises:
            ScheduleError: if the provided ``schedule`` is invalid.
            TQECError: if the provided ``circuit`` contains at least one
                ``stim.CircuitRepeatBlock`` instance.
            TQECError: if the provided ``circuit`` contains at least one
                ``QUBIT_COORDS`` instruction after the first ``TICK`` instruction.

        """
        # Handle each input type explicitly due to type checker
        if isinstance(schedule, int):
            schedule_temp = list(range(schedule, schedule + circuit.num_ticks + 1))
            schedule = Schedule(schedule=schedule_temp)
        elif isinstance(schedule, list) and not isinstance(schedule, Schedule):
            schedule = Schedule(schedule=schedule)

        # Ensure that the provided circuit does not contain any
        # `stim.CircuitRepeatBlock` instance.
        if any(isinstance(inst, stim.CircuitRepeatBlock) for inst in circuit):
            raise ScheduleError(
                "stim.CircuitRepeatBlock instances are not supported in a "
                "ScheduledCircuit instance."
            )
        moments: list[Moment] = list(
            iter_stim_circuit_without_repeat_by_moments(circuit, collected_before_use=True)
        )
        if not moments:
            return ScheduledCircuit.empty()

        if any(m.contains_instruction("QUBIT_COORDS") for m in moments[1:]):
            raise ScheduleError(
                "ScheduledCircuit instance expects the input `stim.Circuit` to "
                "only contain QUBIT_COORDS instructions before the first TICK."
            )
        if qubit_map is None:
            # Here, we know for sure that no qubit is (re)defined in another place
            # than the first moment, so we can directly get qubit coordinates from
            # that moment.
            qubit_map = get_qubit_map(moments[0].circuit)
        # And because we want the cleanest possible moments, we can remove the
        # `QUBIT_COORDS` instructions from the first moment.
        moments[0].remove_all_instructions_inplace(frozenset(["QUBIT_COORDS"]))
        return ScheduledCircuit(moments, schedule, qubit_map, _avoid_checks=True)

    @property
    def schedule(self) -> Schedule:
        """Schedule of the internal moments."""
        return self._schedule

    def get_qubit_coords_definition_preamble(self, shift_to_positive: bool = False) -> stim.Circuit:
        """Get a circuit with only ``QUBIT_COORDS`` instructions.

        Args:
            shift_to_positive: if ``True``, the qubit coordinates are shift such
                that they are all positive. Their relative positioning stays
                unchanged.

        """
        return self._qubit_map.to_circuit(shift_to_positive)

    def get_circuit(
        self, include_qubit_coords: bool = True, shift_to_positive: bool = False
    ) -> stim.Circuit:
        """Build and return the ``stim.Circuit`` instance represented by ``self``.

        Args:
            include_qubit_coords: if ``True``, ``QUBIT_COORDS`` annotations are
                added at the beginning of the returned circuit to declare the
                coordinates of used qubits. Else, no ``QUBIT_COORDS`` annotation
                is added.
            shift_to_positive: if ``True``, the qubit coordinates are shift such
                that they are all positive. Their relative positioning stays
                unchanged.

        Warning:
            The circuit is re-built at each call! Use that function wisely.

        Returns:
            ``stim.Circuit`` instance represented by ``self``.

        """
        ret = stim.Circuit()
        if not self._moments:
            return ret

        # Appending the QUBIT_COORDS instructions first.
        if include_qubit_coords:
            ret += self.get_qubit_coords_definition_preamble(shift_to_positive)

        # Building the actual circuit.
        current_schedule: int = 0
        last_schedule: int = self._schedule[-1]
        for sched, moment in zip(self._schedule, self._moments):
            for _ in range(current_schedule, sched):
                ret.append("TICK", [], [])
                current_schedule += 1
            ret += moment.circuit
            if current_schedule != last_schedule:
                ret.append("TICK", [], [])
                current_schedule += 1

        return ret

    def map_qubit_indices(
        self, qubit_index_map: dict[int, int], inplace: bool = False
    ) -> ScheduledCircuit:
        """Map the qubit **indices** on whom the :class:`ScheduledCircuit` instance is applied.

        Note:
            This method differs from :meth:`~ScheduledCircuit.map_to_qubits`
            because it changes the qubit indices, which requires to iterate the
            whole circuit and change every instruction target.

            This method should be used before combining several
            :class:`ScheduledCircuit` instances that are not aware of each
            other.

        Args:
            qubit_index_map: the map used to modify the qubit targets.
            inplace: if ``True``, perform the modification in place and return
                ``self``. Else, perform the modification in a copy and return
                the copy. Note that the runtime cost of this method should be
                the same independently of the value provided here.

        Returns:
            a modified instance of :class:`ScheduledCircuit` (a copy if
            ``inplace`` is ``True``, else ``self``).

        """
        mapped_final_qubits = QubitMap(
            {qubit_index_map[qi]: q for qi, q in self._qubit_map.items()}
        )
        mapped_moments: list[Moment] = [
            moment.with_mapped_qubit_indices(qubit_index_map) for moment in self._moments
        ]

        if inplace:
            self._qubit_map = mapped_final_qubits
            self._moments = mapped_moments
            return self
        else:
            return ScheduledCircuit(
                mapped_moments, self._schedule, mapped_final_qubits, _avoid_checks=True
            )

    def map_to_qubits(
        self,
        qubit_map: Callable[[GridQubit], GridQubit],
        inplace_qubit_map: bool = False,
    ) -> ScheduledCircuit:
        """Map the qubits the :class:`ScheduledCircuit` instance is applied on.

        Note:
            This method only changes the ``QUBIT_COORDS`` instructions at the
            beginning of the circuit. As such, it never has to iterate on the
            whole quantum circuit and so this method is very efficient.

        Warning:
            The underlying quantum circuit data-structure is never copied (even
            if ``inplace_qubit_map == True``), so the returned instance should be
            used with care, in particular if any method mutating the underlying
            circuit is called on ``self`` or the returned instance after calling
            that method.

        Args:
            qubit_map: the map used to modify the qubits.
            inplace_qubit_map: if True, replaces the qubit map directly in
                ``self`` and return ``self``. Else, create a new instance from
                ``self`` **without copying the underlying moments**, simply
                replacing the qubit map.

        Returns:
            an instance of :class:`ScheduledCircuit` with a new qubit map.

        """
        operand = self if inplace_qubit_map else copy(self)
        operand._qubit_map = operand._qubit_map.with_mapped_qubits(qubit_map)
        return operand

    def __copy__(self) -> ScheduledCircuit:
        return ScheduledCircuit(self._moments, self._schedule, self._qubit_map, _avoid_checks=True)

    def __deepcopy__(self, _: dict[Any, Any]) -> ScheduledCircuit:
        return ScheduledCircuit(
            deepcopy(self._moments),
            deepcopy(self._schedule),
            deepcopy(self._qubit_map),
            _avoid_checks=True,
        )

    @property
    def scheduled_moments(self) -> Iterator[tuple[int, Moment]]:
        """Yields ``stim.Circuit`` instances representing a moment with their computed schedule.

        This property yields all the scheduled moments.

        The yielded elements are sorted with respect to their schedule.

        """
        yield from zip(self._schedule, self._moments)

    @property
    def moments(self) -> Iterator[Moment]:
        """Iterator over the internal moments representing the computation."""
        yield from self._moments

    @property
    def qubits(self) -> frozenset[GridQubit]:
        """Set of qubits on which the circuit is applied."""
        return frozenset(self._qubit_map.qubits)

    def _get_moment_index_by_schedule(self, schedule: int) -> int | None:
        """Get the index of the moment scheduled at the provided schedule.

        Args:
            schedule: the schedule at which the Moment instance should be returned.
                If the schedule is negative, it is considered as an index from
                the end. For example, ``-1`` is the last schedule and ``-2`` is the
                ``last schedule - 1`` (which might not be the second to last
                schedule).

        Returns:
            the index of the :class:`~tqec.circuit.moment.Moment` instance
            scheduled at the provided schedule. If no
            :class:`~tqec.circuit.moment.Moment` instance is scheduled at the
            provided ``schedule``, ``None`` is returned.

        Raises:
            TQECError: if the provided calculated ``schedule`` is negative.

        """
        if not self._schedule:
            return None

        schedule = self._schedule[-1] + 1 + schedule if schedule < 0 else schedule
        if schedule < 0:
            raise TQECError(
                f"Trying to get the index of a Moment instance with a negative schedule {schedule}."
            )
        moment_index = next(
            (i for i, sched in enumerate(self._schedule) if sched == schedule), None
        )
        return moment_index

    def moment_at_schedule(self, schedule: int) -> Moment:
        """Get the :class:`~tqec.circuit.moment.Moment` instance scheduled at the provided schedule.

        Args:
            schedule: the schedule at which the :class:`~tqec.circuit.moment.Moment`
                instance should be returned. If the ``schedule`` is negative, it
                is considered as an index from the end. For example, ``-1`` is the
                last schedule and ``-2`` is the ``last schedule - 1`` (which
                might not be the second to last schedule).

        Raises:
            TQECError: if no moment exist at the provided ``schedule``.

        """
        moment_index = self._get_moment_index_by_schedule(schedule)
        if moment_index is None:
            raise TQECError(f"No Moment instance scheduled at the provided schedule {schedule}.")
        return self._moments[moment_index]

    def append_new_moment(self, moment: Moment) -> None:
        """Schedule the provided :class:`.Moment` instance at the end of the circuit.

        The new schedule will be the last schedule plus one.

        Args:
            moment: the moment to schedule.

        """
        # By default, we cannot assume that self._schedule contains an entry, so
        # we insert at the first moment.
        schedule = Schedule._INITIAL_SCHEDULE
        # If it turns out that we have at least one scheduled moment, then
        # insert just after.
        if self._schedule:
            schedule = self._schedule[-1] + 1
        self.add_to_schedule_index(schedule, moment)

    def reschedule_moment(self, old_schedule: int, new_schedule: int) -> None:
        """Reschedule the moment at ``old_schedule`` to ``new_schedule``.

        Args:
            old_schedule: the schedule at which the moment to reschedule is
                currently scheduled.
            new_schedule: the new schedule at which the moment should be
                scheduled.

        Notes:
            If the old/new schedule is negative, it is considered as an
                index from the end. For example, ``-1`` is the last schedule
                and ``-2`` is the ``last schedule - 1`` (which might not be
                the second to last schedule).

        Raises:
            TQECError: if no moment exist at the provided ``old_schedule``.

        """
        moment_index = self._get_moment_index_by_schedule(old_schedule)
        if moment_index is None:
            raise TQECError(
                f"No Moment instance scheduled at the provided schedule {old_schedule}."
            )
        moment = self._moments.pop(moment_index)
        self._schedule.schedule.pop(moment_index)
        self.add_to_schedule_index(new_schedule, moment)

    def add_to_schedule_index(self, schedule: int, moment: Moment) -> None:
        """Add the operations contained in the provided ``moment`` to the provided schedule.

        Args:
            schedule: schedule at which operations in ``moment`` should be added.
                If the schedule is negative, it is considered as an index from
                the end. For example, ``-1`` is the last schedule and ``-2`` is the
                ``last schedule - 1`` (which might not be the second to last
                schedule).
            moment: operations that should be added.

        """
        moment_index = self._get_moment_index_by_schedule(schedule)
        if moment_index is None:
            # We have to insert a new schedule and a new Moment.
            insertion_index = bisect.bisect_left(self._schedule, schedule)
            self._schedule.insert(insertion_index, schedule)
            self._moments.insert(insertion_index, moment)
        else:
            # Else, the schedule already exists, in which case we just need to
            # add the operations to an existing moment. Note that this might
            # fail if two instructions overlap.
            self._moments[moment_index] += moment

    def append_observable(self, index: int, targets: Sequence[stim.GateTarget]) -> None:
        """Append an ``OBSERVABLE_INCLUDE`` instruction to the last moment.

        Args:
            index: index of the observable to append measurement records to.
            targets: measurement records forming (part of) the observable.

        """
        self.append_annotation(stim.CircuitInstruction("OBSERVABLE_INCLUDE", targets, [index]))

    def append_annotation(self, instruction: stim.CircuitInstruction) -> None:
        """Append an annotation to the last moment.

        Args:
            instruction: an annotation that will be added to the last moment of
                ``self``.

        Raises:
            TQECError: if the provided instruction is not an annotation.

        """
        if not is_annotation_instruction(instruction):
            raise TQECError(
                "The provided instruction is not an annotation, which is "
                "disallowed by the append_annotation method."
            )
        self._moments[-1].append_annotation(instruction)

    @property
    def num_measurements(self) -> int:
        """Number of measurements in the represented computation."""
        return sum(m.num_measurements for m in self._moments)

    def filter_by_qubits(self, qubits_to_keep: Iterable[GridQubit]) -> ScheduledCircuit:
        """Filter the circuit to keep only the instructions that are applied on the provided qubits.

        If an instruction is applied on a qubit that is not in the provided list, it is removed.

        After filtering, the empty moments as well as the corresponding schedules
        are removed.

        Args:
            qubits_to_keep: the qubits to keep in the circuit.

        Returns:
            a new instance of :class:`ScheduledCircuit` with the filtered
            circuit and schedules.

        """
        qubits_indices_to_keep = frozenset(
            self._qubit_map.q2i[q] for q in qubits_to_keep if q in self.qubits
        )
        filtered_moments: list[Moment] = []
        filtered_schedule: list[int] = []
        for schedule, moment in self.scheduled_moments:
            filtered_moment = moment.filter_by_qubits(qubits_indices_to_keep)
            if filtered_moment.is_empty:
                continue
            filtered_moments.append(filtered_moment)
            filtered_schedule.append(schedule)
        qubit_map = self._qubit_map.filter_by_qubits(qubits_to_keep)
        filtered_circuit = ScheduledCircuit(
            filtered_moments, Schedule(filtered_schedule), qubit_map, _avoid_checks=True
        )
        # The qubit indices may not be contiguous anymore, so we need to remap them.
        indices_map = {oi: ni for ni, oi in enumerate(qubit_map.indices)}
        return filtered_circuit.map_qubit_indices(indices_map)

    @property
    def qubit_map(self) -> QubitMap:
        """Qubit map of the circuit."""
        return self._qubit_map

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ScheduledCircuit)
            and self._schedule == other._schedule
            and self._qubit_map == other._qubit_map
            and self._moments == other._moments
        )

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    def is_empty(self) -> bool:
        """Return ``True`` if ``self`` represents an empty circuit.

        Warning:
            an empty circuit is defined as a circuit containing no instructions.
            ``QUBIT_COORDINATES`` instructions are ignored by this method (i.e., a circuit with only
            ``QUBIT_COORDINATES`` instructions is considered empty), but other annotations are taken
            into account (i.e., a circuit with a ``DETECTOR`` or ``SHIFT_COORDS`` instruction will
            not be considered empty).

        """
        return all(m.is_empty for m in self._moments)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the circuit.

        Returns:
            a dictionary with the keys ``moments``, ``schedule`` and
            ``qubit_map`` and their corresponding values.

        """
        return {
            "moments": [m.to_dict() for m in self._moments],
            "schedule": self._schedule.schedule,
            "qubit_map": self._qubit_map.to_dict(),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> ScheduledCircuit:
        """Return a circuit from its dictionary representation.

        Args:
            data: dictionary with the keys ``moments``, ``schedule`` and
                ``qubit_map``.

        """
        moments = [Moment.from_dict(m) for m in data["moments"]]
        schedule = Schedule(data["schedule"])
        qubit_map = QubitMap.from_dict(data["qubit_map"])
        return ScheduledCircuit(moments, schedule, qubit_map, _avoid_checks=True)



from __future__ import annotations

import functools
import operator
from collections.abc import Callable, Iterator
from dataclasses import dataclass

import stim

from tqec.circuit.moment import Moment
from tqec.circuit.schedule.manipulation import merge_instructions


@dataclass
class RepeatedMoments:
    repetitions: int
    moments: list[Moment | RepeatedMoments]


def iter_stim_circuit_by_moments(
    circuit: stim.Circuit, collected_before_use: bool = True
) -> Iterator[Moment | RepeatedMoments]:
    """Return an iterator over the moments of the provided ``circuit``.

    A moment is defined as all the operations between two ``TICK`` instructions. A moment can be
    empty (i.e., contains no instruction).

    Args:
        circuit: quantum circuit to extract moments from.
        collected_before_use: if ``False``, each call to ``next`` on the returned iterator will
            erase the previously returned circuit and replace it with a new one. Should only be set
            to ``False`` when iterating **AND** analysing moments one after the other, without the
            need to collect all the moments before.

    Returns:
        all the moments appearing in the provided ``circuit``, in apparition order. The quantum
        circuits representing each moment does **NOT** contain any ``TICK`` instruction.

    """
    copy_func: Callable[[stim.Circuit], stim.Circuit] = (
        (lambda c: c.copy()) if collected_before_use else (lambda c: c)
    )
    cur_moment = stim.Circuit()
    for inst in circuit:
        if isinstance(inst, stim.CircuitRepeatBlock):
            yield Moment(copy_func(cur_moment))
            cur_moment.clear()
            yield RepeatedMoments(
                inst.repeat_count,
                list(iter_stim_circuit_by_moments(inst.body_copy(), collected_before_use=True)),
            )
        elif inst.name == "TICK":
            yield Moment(copy_func(cur_moment))
            cur_moment.clear()
        else:
            cur_moment.append(inst)
    # No need to copy the last moment
    yield Moment(cur_moment)


def collect_moments(moments: list[Moment | RepeatedMoments]) -> stim.Circuit:
    """Merge the provided ``moments`` into a ``stim.Circuit``."""
    if not moments:
        return stim.Circuit()

    ret = stim.Circuit()
    if isinstance(moments[0], Moment):
        ret += moments[0].circuit
    else:
        repeated_block = stim.Circuit("TICK") + collect_moments(moments[0].moments)
        ret.append(stim.CircuitRepeatBlock(moments[0].repetitions, repeated_block))

    for i in range(1, len(moments)):
        moment = moments[i]
        if isinstance(moment, Moment):
            ret.append(stim.CircuitInstruction("TICK"))
            ret += moment.circuit
        else:
            repeated_block = stim.Circuit("TICK") + collect_moments(moment.moments)
            ret.append(stim.CircuitRepeatBlock(moment.repetitions, repeated_block))
    return ret


def merge_moments(lhs: Moment, rhs: Moment) -> Moment:
    """Merge two :class:`.Moment` instances into a single instance.

    Raises:
        TQECError if the resulting :class:`.Moment` instance is invalid, for example if it
            contains two operations using the same qubit at the same time step.

    """
    instructions = merge_instructions(list(lhs.instructions) + list(rhs.instructions))
    circuit = stim.Circuit()
    for inst in instructions:
        circuit.append(
            inst.name,
            functools.reduce(
                operator.iadd,
                sorted(
                    inst.target_groups(),
                    key=lambda group: tuple(t.value for t in group),
                ),
                [],
            ),
            inst.gate_args_copy(),
        )
    return Moment(circuit)



"""Defines :class:`.ScheduledCircuit` used to represent quantum circuits in the :mod:`tqec` package.

This module defines one of the main class of the :mod:`tqec` package,
:class:`~.circuit.ScheduledCircuit`, and several functions to modify or create
instances:

- :func:`~.manipulation.merge_scheduled_circuits` that takes several instances of
  :class:`~.circuit.ScheduledCircuit` and merge them into one instance,
- :func:`~.manipulation.relabel_circuits_qubit_indices` that takes several
  instances of :class:`~.circuit.ScheduledCircuit` and relabel their qubits to
  avoid any index clash when merging,
- :func:`~.manipulation.remove_duplicate_instructions` that removes some
  instructions when they are duplicated in a :class:`~tqec.circuit.moment.Moment`
  instance.

"""

from .circuit import ScheduledCircuit as ScheduledCircuit
from .exception import ScheduleError as ScheduleError
from .manipulation import merge_scheduled_circuits as merge_scheduled_circuits
from .manipulation import (
    relabel_circuits_qubit_indices as relabel_circuits_qubit_indices,
)
from .manipulation import remove_duplicate_instructions as remove_duplicate_instructions
from .schedule import Schedule as Schedule



from __future__ import annotations

from enum import Enum, auto
from typing import cast

import svg
from typing_extensions import override

from tqec.utils.enums import Basis
from tqec.visualisation.computation.plaquette.base import (
    PlaquetteCorner,
    SVGPlaquetteDrawer,
    lerp,
    svg_path_enclosing_points,
)
from tqec.visualisation.configuration import DrawerConfiguration
from tqec.visualisation.exception import TQECDrawingError


class ExtendedPlaquettePosition(Enum):
    UP = "UP"
    DOWN = "DOWN"

    def flip(self) -> ExtendedPlaquettePosition:
        """Return the opposite direction."""
        match self:
            case ExtendedPlaquettePosition.UP:
                return ExtendedPlaquettePosition.DOWN
            case ExtendedPlaquettePosition.DOWN:
                return ExtendedPlaquettePosition.UP
            # wildcard entry added as it was flagged by ty
            case _:
                raise ValueError("Unexpected input provided.")


class ExtendedPlaquetteType(Enum):
    BULK = auto()
    LEFT_WITH_ARM = auto()
    LEFT_WITHOUT_ARM = auto()
    RIGHT_WITH_ARM = auto()
    RIGHT_WITHOUT_ARM = auto()


class ExtendedPlaquetteDrawer(SVGPlaquetteDrawer):
    def __init__(
        self,
        plaquette_type: ExtendedPlaquetteType,
        position: ExtendedPlaquettePosition,
        basis: Basis,
        schedule: tuple[int, int, int, int],
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> None:
        """SVG plaquette drawer for extended plaquettes."""
        super().__init__()
        self._plaquette_type = plaquette_type
        self._position = position
        self._basis = basis
        self._schedule = (
            schedule[0:2] if position == ExtendedPlaquettePosition.UP else schedule[2:4]
        )
        self._reset = reset
        self._measurement = measurement

    @override
    def draw(
        self,
        id: str,
        show_interaction_order: bool = True,
        show_hook_errors: bool = True,
        show_data_qubit_reset_measurements: bool = True,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        # Build iteratively the different layers that we need to represent the
        # plaquette.
        shape_path = self.get_plaquette_shape_path(configuration)
        layers: list[svg.Element] = []
        # If the plaquette has a uniform color, the shape include that color with
        # its "fill" attribute and so we do not need to clip. If that is not the
        # case, the filling colours will be added with get_fill_layer, but it is
        # simpler for the implementation to clip.
        layers.append(shape_path)
        if show_data_qubit_reset_measurements:
            layers.append(self.get_data_qubit_reset_measurements_layers(configuration))
        if show_hook_errors:
            if (line := self.get_hook_error_line(configuration)) is not None:
                layers.append(line)
        if show_interaction_order:
            layers.extend(self.get_interaction_order_text(configuration))
        return svg.G(id=id, elements=layers)

    @staticmethod
    def _get_extended_plaquette_square_shape(
        position: ExtendedPlaquettePosition,
        plaquette_type: ExtendedPlaquetteType,
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.G:
        tl, tr, bl, br = SVGPlaquetteDrawer._CORNERS
        if plaquette_type == ExtendedPlaquetteType.LEFT_WITHOUT_ARM:
            tl, bl = (tl + tr) / 2, (bl + br) / 2
        elif plaquette_type == ExtendedPlaquetteType.RIGHT_WITHOUT_ARM:
            tr, br = (tl + tr) / 2, (bl + br) / 2

        up_stroke_color, down_stroke_color = (
            configuration.stroke_color,
            configuration.thin_stroke_color,
        )
        if position == ExtendedPlaquettePosition.DOWN:
            up_stroke_color, down_stroke_color = down_stroke_color, up_stroke_color
        return svg.G(
            elements=[
                # Filling
                svg.Rect(
                    x=tl.real,
                    y=tl.imag,
                    width=(br.real - tl.real),
                    height=(br.imag - tl.imag),
                    fill=fill,
                ),
                # UP
                svg.Line(
                    x1=tl.real,
                    y1=tl.imag,
                    x2=tr.real,
                    y2=tr.imag,
                    stroke=up_stroke_color,
                    stroke_width=configuration.stroke_width,
                ),
                # DOWN
                svg.Line(
                    x1=bl.real,
                    y1=bl.imag,
                    x2=br.real,
                    y2=br.imag,
                    stroke=down_stroke_color,
                    stroke_width=configuration.stroke_width,
                ),
                # LEFT
                svg.Line(
                    x1=tl.real,
                    y1=tl.imag,
                    x2=bl.real,
                    y2=bl.imag,
                    stroke=configuration.stroke_color,
                    stroke_width=configuration.stroke_width,
                ),
                # RIGHT
                svg.Line(
                    x1=tr.real,
                    y1=tr.imag,
                    x2=br.real,
                    y2=br.imag,
                    stroke=configuration.stroke_color,
                    stroke_width=configuration.stroke_width,
                ),
            ]
        )

    @staticmethod
    def _get_weight_three_extended_plaquette_shape(
        position: ExtendedPlaquettePosition,
        plaquette_type: ExtendedPlaquetteType,
        fill: str = "none",
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        if position == ExtendedPlaquettePosition.DOWN:
            return svg.G()
        _, tr, bl, br = SVGPlaquetteDrawer._CORNERS
        bl += 1j
        br += 1j
        center = 0.5 + 1j
        side_length = configuration.plaquette_overflow_lerp_coefficient

        vs = [bl, bl - side_length * 1j, tr - side_length, tr, br]
        if plaquette_type == ExtendedPlaquetteType.RIGHT_WITH_ARM:
            vs = [2 * center - v for v in vs]
        return svg_path_enclosing_points(vs, fill, configuration)

    def get_plaquette_shape_path(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        """Return the plaquette shape, uniformly filled with the appropriate color.

        Args:
            configuration: drawing configuration.

        Returns:
            The plaquette shape. It is filled with the appropriate colour corresponding to
            the value of ``self._basis``.

        """
        fill = SVGPlaquetteDrawer.get_colour(self._basis)
        match self._plaquette_type:
            case (
                ExtendedPlaquetteType.BULK
                | ExtendedPlaquetteType.LEFT_WITHOUT_ARM
                | ExtendedPlaquetteType.RIGHT_WITHOUT_ARM
            ):
                return ExtendedPlaquetteDrawer._get_extended_plaquette_square_shape(
                    self._position, self._plaquette_type, fill, configuration
                )
            case ExtendedPlaquetteType.LEFT_WITH_ARM | ExtendedPlaquetteType.RIGHT_WITH_ARM:
                return ExtendedPlaquetteDrawer._get_weight_three_extended_plaquette_shape(
                    self._position, self._plaquette_type, fill, configuration
                )
            # wildcard entry added as it was flagged by ty
            case _:
                raise ValueError("Unsupported input provided.")

    def get_interaction_order_text(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> list[svg.Text]:
        """Return a SVG element containing data-qubit interaction orders as text.

        This function returns one SVG element per non-empty corners, each containing a text element
        with the time slice at which a 2-qubit operation is applied on the corner qubit.

        Args:
            configuration: drawing configuration.

        Returns:
            One SVG element per non-empty corners, each containing a text
            element with the time slice at which a 2-qubit operation is applied
            on the corner qubit.

        """
        interaction_order_texts: list[svg.Text] = []
        tl, tr, bl, br = SVGPlaquetteDrawer._CORNERS
        s1, s2 = self._schedule
        data_corners: list[complex]
        schedules: list[int]
        match self._plaquette_type:
            case ExtendedPlaquetteType.BULK:
                data_corners = (
                    [tl, tr] if self._position == ExtendedPlaquettePosition.UP else [bl, br]
                )
                schedules = [s1, s2]
            case ExtendedPlaquetteType.LEFT_WITH_ARM:
                data_corners = [tr] if self._position == ExtendedPlaquettePosition.UP else [bl, br]
                schedules = [s2] if self._position == ExtendedPlaquettePosition.UP else [s1, s2]
            case ExtendedPlaquetteType.RIGHT_WITH_ARM:
                data_corners = [tl, tr] if self._position == ExtendedPlaquettePosition.UP else [bl]
                schedules = [s1, s2] if self._position == ExtendedPlaquettePosition.UP else [s1]
            case ExtendedPlaquetteType.LEFT_WITHOUT_ARM:
                data_corners = [tr] if self._position == ExtendedPlaquettePosition.UP else [br]
                schedules = [s2]
            case ExtendedPlaquetteType.RIGHT_WITHOUT_ARM:
                data_corners = [tl] if self._position == ExtendedPlaquettePosition.UP else [bl]
                schedules = [s1]

        for corner, schedule in zip(data_corners, schedules):
            if not schedule:
                continue
            text_position = lerp(
                SVGPlaquetteDrawer._CENTER_COORDINATE, corner, configuration.text_lerp_coefficient
            )
            interaction_order_texts.append(
                svg.Text(
                    x=text_position.real,
                    y=text_position.imag,
                    fill="black",
                    font_size=0.1,
                    text_anchor="middle",
                    dominant_baseline="central",
                    text=str(schedule),
                )
            )
        return interaction_order_texts

    def get_hook_error_line(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Line | None:
        """Return a SVG line showing the direction of the hook error.

        Args:
            configuration: drawing configuration.

        Returns:
            A SVG line showing the direction of the hook error, or ``None`` if
            there is no hook error.

        """
        # No hook error for these.
        if self._plaquette_type in [
            ExtendedPlaquetteType.LEFT_WITHOUT_ARM,
            ExtendedPlaquetteType.RIGHT_WITHOUT_ARM,
        ]:
            return None
        if (
            self._plaquette_type == ExtendedPlaquetteType.LEFT_WITH_ARM
            and self._position == ExtendedPlaquettePosition.UP
        ) or (
            self._plaquette_type == ExtendedPlaquetteType.RIGHT_WITH_ARM
            and self._position == ExtendedPlaquettePosition.DOWN
        ):
            return None

        if (
            self._plaquette_type == ExtendedPlaquetteType.BULK
            and self._position == ExtendedPlaquettePosition.UP
        ):
            return None

        tl, tr, bl, br = SVGPlaquetteDrawer._CORNERS
        c1, c2 = (tl, tr) if self._position == ExtendedPlaquettePosition.UP else (bl, br)
        f = configuration.hook_error_line_lerp_coefficient
        a = lerp(SVGPlaquetteDrawer._CENTER_COORDINATE, c1, f)
        b = lerp(SVGPlaquetteDrawer._CENTER_COORDINATE, c2, f)
        return svg.Line(
            x1=a.real,
            x2=b.real,
            y1=a.imag,
            y2=b.imag,
            stroke=configuration.stroke_color,
            stroke_width=configuration.stroke_width,
        )

    def get_data_qubit_reset_measurements_layers(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.G:
        """Return a SVG layer containing a representation of data-qubit resets/measurements.

        Args:
            configuration: drawing configuration.

        Returns:
            a SVG element containing all the reset or measurements on the
            plaquette drawn by ``self``.

        """
        if self._reset is None and self._measurement is None:
            return svg.G()
        if self._reset is not None and self._measurement is not None:
            raise TQECDrawingError("Cannot draw both reset and measurement.")
        basis = self._reset if self._reset is not None else cast(Basis, self._measurement)
        fill = SVGPlaquetteDrawer.get_colour(basis)
        match self._plaquette_type:
            case ExtendedPlaquetteType.BULK:
                places = (
                    [PlaquetteCorner.TOP_LEFT, PlaquetteCorner.TOP_RIGHT]
                    if self._position == ExtendedPlaquettePosition.UP
                    else [PlaquetteCorner.BOTTOM_LEFT, PlaquetteCorner.BOTTOM_RIGHT]
                )
            case ExtendedPlaquetteType.LEFT_WITH_ARM:
                places = (
                    [PlaquetteCorner.TOP_RIGHT]
                    if self._position == ExtendedPlaquettePosition.UP
                    else [PlaquetteCorner.BOTTOM_LEFT, PlaquetteCorner.BOTTOM_RIGHT]
                )
            case ExtendedPlaquetteType.RIGHT_WITH_ARM:
                places = (
                    [PlaquetteCorner.TOP_LEFT, PlaquetteCorner.TOP_RIGHT]
                    if self._position == ExtendedPlaquettePosition.UP
                    else [PlaquetteCorner.BOTTOM_LEFT]
                )
            case ExtendedPlaquetteType.LEFT_WITHOUT_ARM:
                places = (
                    [PlaquetteCorner.TOP_RIGHT]
                    if self._position == ExtendedPlaquettePosition.UP
                    else [PlaquetteCorner.BOTTOM_RIGHT]
                )
            case ExtendedPlaquetteType.RIGHT_WITHOUT_ARM:
                places = (
                    [PlaquetteCorner.TOP_LEFT]
                    if self._position == ExtendedPlaquettePosition.UP
                    else [PlaquetteCorner.BOTTOM_LEFT]
                )
        reset_measurement_svgs: list[svg.Element] = []
        for place in places:
            corner_coords = self.get_corner_coordinates(place)
            reset_measurement_svgs.append(
                svg.G(
                    elements=[
                        self.get_reset_shape(place, fill, configuration)
                        if self._reset is not None
                        else self.get_measurement_shape(place, fill, configuration)
                    ],
                    transform=[svg.Translate(x=corner_coords.real, y=corner_coords.imag)],
                )
            )
        return svg.G(elements=reset_measurement_svgs)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from tqec.compile.observables.fixed_boundary_builder import (
    FIXED_BOUNDARY_OBSERVABLE_BUILDER,
)
from tqec.compile.observables.fixed_bulk_builder import FIXED_BULK_OBSERVABLE_BUILDER
from tqec.compile.specs.library.fixed_boundary import (
    FIXED_BOUNDARY_CUBE_BUILDER,
    FIXED_BOUNDARY_PIPE_BUILDER,
)
from tqec.compile.specs.library.fixed_bulk import (
    FIXED_BULK_CUBE_BUILDER,
    FIXED_BULK_PIPE_BUILDER,
)

if TYPE_CHECKING:
    from tqec.compile.observables.builder import ObservableBuilder
    from tqec.compile.specs.base import CubeBuilder, PipeBuilder


@dataclass(frozen=True)
class ConventionTriplet:
    """Stores the 3 builders needed to implement a new convention.

    In order to implement a new way of generating plaquettes and implementing
    blocks, a new :class:`Convention` should be created. This involves
    implementing the interfaces for each of the 3 attributes below.

    """

    cube_builder: CubeBuilder
    pipe_builder: PipeBuilder
    observable_builder: ObservableBuilder


@dataclass(frozen=True)
class Convention:
    """Represent a convention to implement blocks."""

    name: str
    triplet: ConventionTriplet

    def __str__(self) -> str:
        return self.name  # pragma: no cover


FIXED_BULK_CONVENTION = Convention(
    "fixed_bulk",
    ConventionTriplet(
        FIXED_BULK_CUBE_BUILDER, FIXED_BULK_PIPE_BUILDER, FIXED_BULK_OBSERVABLE_BUILDER
    ),
)

FIXED_BOUNDARY_CONVENTION = Convention(
    "fixed_boundary",
    ConventionTriplet(
        FIXED_BOUNDARY_CUBE_BUILDER,
        FIXED_BOUNDARY_PIPE_BUILDER,
        FIXED_BOUNDARY_OBSERVABLE_BUILDER,
    ),
)

ALL_CONVENTIONS = {conv.name: conv for conv in [FIXED_BULK_CONVENTION, FIXED_BOUNDARY_CONVENTION]}



"""Provides functions to pretty-print :class:`~tqec.templates.base.Template` instances."""

from __future__ import annotations

import typing as ty

import numpy
import numpy.typing as npt

from tqec.templates.base import Template


def display_template(
    template: Template, k: int, plaquette_indices: ty.Sequence[int] | None = None
) -> None:
    """Display a template instance with ASCII output.

    Args:
        template: the Template instance to display.
        k: scaling factor.
        plaquette_indices: the plaquette indices that are forwarded to the call
            to `template.instantiate` to get the actual template representation.
            If None, default to ``range(1, template.expected_plaquettes_number + 1)``.

    """
    if plaquette_indices is None:
        plaquette_indices = tuple(range(1, template.expected_plaquettes_number + 1))
    display_template_from_instantiation(template.instantiate(k, plaquette_indices))


def display_template_from_instantiation(
    instantiation: npt.NDArray[numpy.int_],
) -> None:  # pragma: no cover
    """Display an array representing a template instantiation with ASCII output.

    Args:
        instantiation: the integer array obtained from the `Template.instantiate`
            method.

    """
    print(get_template_representation_from_instantiation(instantiation))


def get_template_representation_from_instantiation(
    instantiation: npt.NDArray[numpy.int_],
) -> str:
    """Get the pretty-string representation of a template instantiation."""
    max_integer = numpy.max(instantiation)
    pad = len(str(max_integer)) + 1
    return "\n".join(
        " ".join((str(num) if num != 0 else ".").rjust(pad) for num in line)
        for line in instantiation
    )



"""Read and write simulation data from and to csv files."""

import io
from collections.abc import Iterable
from pathlib import Path
from typing import Literal, TextIO

import sinter

from tqec.utils.exceptions import TQECError


def write_sinter_stats_to_csv(
    filepath: str | Path,
    stats: Iterable[sinter.TaskStats],
    if_file_exists: Literal["merge", "overwrite", "raise"] = "raise",
) -> None:
    """Write simulation data to csv file.

    This is a convenience function because ``sinter`` does not expose this functionality. To read
    simulation data use ``sinter.read_stats_from_csv_files``.

    Args:
        filepath: The file to be written to.
        stats: List of simulation results. Each element corresponds to a line in a csv file.
        if_file_exists: How to react if ``filepath`` points to an existing file.
            - raise: Raise a ``TQECError`` if the file exists.
            - overwrite: Effectively delete the file and pretend none exists.
            - merge: Read the file and try to match entries with the simulation data from ``stats``
                by strong id. In case of a match merge the entries.

    """
    filepath = Path(filepath)

    if filepath.exists():
        if if_file_exists == "raise":
            raise TQECError(f"File '{filepath}' already exists.")
        elif if_file_exists == "overwrite":
            # Overwrite happens later automatically.
            pass
        elif if_file_exists == "merge":
            # Writing stats to a file is necessary because sinter does not expose merging via API.
            # We can only indirectly merge by reading in two files.
            in_memory_file = io.StringIO()
            _write_sinter_stats(in_memory_file, stats)
            _ = in_memory_file.seek(0)
            stats = sinter.read_stats_from_csv_files(filepath, in_memory_file)

    with open(filepath, "w") as f:
        _write_sinter_stats(f, stats)


def _write_sinter_stats(file: TextIO, stats: Iterable[sinter.TaskStats]) -> None:
    _ = file.write(sinter.CSV_HEADER + "\n")
    for sample in stats:
        _ = file.write(sample.to_csv_line() + "\n")



"""Block graph that represents a CZ gate."""

import stim

from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import ZXCube
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Position3D


def cz(support_flows: str | list[str] | None = None) -> BlockGraph:
    """Create a block graph representing the logical CZ gate.

    By default, the Hadamard edge in the CZ gate will be horizontal. If you
    want to use vertical Hadamard edges, you can rotate the graph by 90 degrees
    by calling :py:meth:`~tqec.computation.block_graph.BlockGraph.rotate`.

    Args:
        support_flows: The stabilizer flow supported by the logical CZ gate. It
            determines the node kind at the ports of the graph. It can be either
            a single string like "XI -> XZ" or an iterable of strings like
            ["XI -> XZ", "IZ -> IZ"] or None. If multiple flows are provided,
            a valid graph will be computed or an error will be raised if no
            valid graph can be found. If None, the graph will have open ports.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance representing the
        logical CZ gate.

    Raises:
        TQECError: If there is Y Pauli operator in the stabilizer flows, or
            if provided stabilizer flows are not valid for the CZ gate, or
            if no valid graph can be found for the given stabilizer flows.

    """
    g = BlockGraph("Logical CZ")
    nodes = [
        (Position3D(0, 0, 0), "P", "In_1"),
        (Position3D(0, 0, 1), "XZX", ""),
        (Position3D(0, 0, 2), "P", "Out_1"),
        (Position3D(1, 0, 1), "XXZ", ""),
        (Position3D(1, -1, 1), "P", "In_2"),
        (Position3D(1, 1, 1), "P", "Out_2"),
    ]
    for pos, kind, label in nodes:
        g.add_cube(pos, kind, label)

    pipes = [(0, 1), (1, 2), (1, 3), (3, 4), (3, 5)]
    for p0, p1 in pipes:
        g.add_pipe(nodes[p0][0], nodes[p1][0])

    if support_flows is not None:
        flows = (
            [f.upper() for f in support_flows]
            if isinstance(support_flows, list)
            else [support_flows.upper()]
        )
        resolved_ports = _resolve_ports(flows)
        g.fill_ports(dict(zip(["In_1", "In_2", "Out_1", "Out_2"], resolved_ports)))
    return g


def _resolve_ports(flows: list[str]) -> list[ZXCube]:
    if any("Y" in f for f in flows):
        raise TQECError("Y basis initialization/measurements are not supported yet.")

    stim_flows = [stim.Flow(f) for f in flows]
    cz = stim.Circuit("CZ 0 1")
    for f in stim_flows:
        if not cz.has_flow(f):
            raise TQECError(f"{f} is not a valid flow for the CZ gate.")

    ports = ["_"] * 4
    for f in stim_flows:
        for i, p in enumerate(str(f.input_copy() + f.output_copy())[1:]):
            if ports[i] == "_":
                ports[i] = p
            elif p != "_" and p != ports[i]:
                raise TQECError(f"Port {i} fails to support both {ports[i]} and {p} observable.")
    # If there are left "I" in the ports, we choose fill them with "Z" ("X" should also work).
    for i in range(4):
        if ports[i] == "_":
            ports[i] = "Z"

    def _vkind(p: str) -> ZXCube:
        return ZXCube.from_str(f"XZ{p}")

    def _hkind(p: str) -> ZXCube:
        return ZXCube.from_str(f"X{p}Z")

    return [_vkind(ports[0]), _hkind(ports[1]), _vkind(ports[2]), _hkind(ports[3])]



import stim


def add_tick_coordinate_to_detectors(circuit: stim.Circuit) -> stim.Circuit:
    """Add a coordinate containing the current moment index to each detector."""
    ret = stim.Circuit()
    num_ticks: int = 0
    for instruction in circuit.flattened():
        assert not isinstance(instruction, stim.CircuitRepeatBlock)
        assert instruction.name != "SHIFT_COORDS"
        if instruction.name == "TICK":
            num_ticks += 1
            ret.append(instruction)
        elif instruction.name == "DETECTOR":
            ret.append(
                stim.CircuitInstruction(
                    instruction.name,
                    instruction.targets_copy(),
                    [*instruction.gate_args_copy(), num_ticks],
                )
            )
        else:
            ret.append(instruction)
    return ret



"""Synthesis of :py:class:`~tqec.computation.BlockGraph` instance from PyZX graph."""

from tqec.interop.pyzx.synthesis.positioned import (
    positioned_block_synthesis as positioned_block_synthesis,
)
from tqec.interop.pyzx.synthesis.strategy import (
    SynthesisStrategy as SynthesisStrategy,
)
from tqec.interop.pyzx.synthesis.strategy import (
    block_synthesis as block_synthesis,
)



import functools
from copy import deepcopy
from typing import Final, Literal

import stim
from typing_extensions import override

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.plaquette.constants import MEASUREMENT_SCHEDULE
from tqec.plaquette.debug import PlaquetteDebugInformation
from tqec.plaquette.plaquette import Plaquette
from tqec.plaquette.qubit import PlaquetteQubits, SquarePlaquetteQubits
from tqec.plaquette.rpng import ExtendedBasis, PauliBasis, RPNGDescription
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.utils.exceptions import TQECError
from tqec.utils.instructions import (
    MEASUREMENT_INSTRUCTION_NAMES,
    RESET_INSTRUCTION_NAMES,
)


class DefaultRPNGTranslator(RPNGTranslator):
    """Default implementation of the RPNGTranslator interface.

    The plaquettes returned have the following properties:

    - the syndrome qubit is always reset in the ``X``-basis,
    - the syndrome qubit is always measured in the ``X``-basis,
    - the syndrome qubit is always the control of the 2-qubit gates used,
    - the 2-qubit gate used is always a ``Z``-controlled Pauli gate,
    - resets (and potentially hadamards) are always scheduled at timestep ``0``,
    - 2-qubit gates are always scheduled at timesteps in ``[1, 5]``,
    - measurements (and potentially hadamards) are always scheduled at timestep
      ``DefaultRPNGTranslator.MEASUREMENT_SCHEDULE`` that is currently equal to
      ``6``,
    - resets and measurements are always ordered from their basis (first ``X``,
      then ``Y``, and finally ``Z``),
    - hadamard gates are always after resets and measurements,
    - targets of reset, measurement and hadamard are always ordered.

    """

    QUBITS: Final[PlaquetteQubits] = SquarePlaquetteQubits()

    @staticmethod
    def _add_extended_basis_operation(
        circuit: stim.Circuit,
        op: Literal["R", "M"],
        timestep_operations: dict[ExtendedBasis, list[int]],
    ) -> None:
        for basis in ExtendedBasis:
            if basis not in timestep_operations:
                continue
            targets = sorted(timestep_operations[basis])
            match basis:
                case ExtendedBasis.H:
                    circuit.append("H", targets, [])
                case ExtendedBasis.X | ExtendedBasis.Y | ExtendedBasis.Z:
                    circuit.append(f"{op}{basis.value.upper()}", targets, [])

    @override
    def translate(self, rpng_description: RPNGDescription) -> Plaquette:
        return self._translate_impl(rpng_description)

    @functools.lru_cache(maxsize=1024)
    def _translate_impl(self, rpng_description: RPNGDescription) -> Plaquette:
        # The current RPNG notation is very much tied to the qubit arrangement
        # in SquarePlaquetteQubits, hence the explicit value here.
        qubits: PlaquetteQubits = deepcopy(DefaultRPNGTranslator.QUBITS)

        data_qubit_indices = list(qubits.data_qubits_indices)
        if len(data_qubit_indices) != 4:
            raise TQECError("Expected 4 data-qubits, got", len(data_qubit_indices))
        used_data_qubit_indices: set[int] = set()
        syndrome_qubit_indices = list(qubits.syndrome_qubits_indices)
        if len(syndrome_qubit_indices) != 1:
            raise TQECError("Expected 1 syndrome qubit, got", len(syndrome_qubit_indices))
        syndrome_qubit_index = syndrome_qubit_indices[0]

        # Handling syndrome qubit reset/measurement
        reset_timestep_operations: dict[ExtendedBasis, list[int]] = {}
        meas_timestep_operations: dict[ExtendedBasis, list[int]] = {}
        if (r := rpng_description.ancilla.r) is not None:
            reset_timestep_operations[r.to_extended_basis()] = [syndrome_qubit_index]
        if (g := rpng_description.ancilla.g) is not None:
            meas_timestep_operations[g.to_extended_basis()] = [syndrome_qubit_index]
        # Handling data-qubits
        entangling_operations: list[tuple[PauliBasis, int] | None] = [
            None for _ in range(MEASUREMENT_SCHEDULE - 1)
        ]
        for qi, rpng in enumerate(rpng_description.corners):
            dqi = data_qubit_indices[qi]
            if rpng.r is not None:
                reset_timestep_operations.setdefault(rpng.r, []).append(dqi)
                used_data_qubit_indices.add(dqi)
            if rpng.g is not None:
                meas_timestep_operations.setdefault(rpng.g, []).append(dqi)
                used_data_qubit_indices.add(dqi)
            if rpng.p is not None and rpng.n is not None:
                entangling_operations[rpng.n - 1] = (rpng.p, dqi)
                used_data_qubit_indices.add(dqi)

        circuit = stim.Circuit()
        schedule: list[int] = [0]
        # Add reset operations
        self._add_extended_basis_operation(circuit, "R", reset_timestep_operations)
        circuit.append("TICK", [], [])

        # Add entangling gates
        for sched, entangling_operation in enumerate(entangling_operations):
            if entangling_operation is None:
                continue
            p, data_qubit = entangling_operation
            circuit.append(f"C{p.value.upper()}", [syndrome_qubit_index, data_qubit], [])
            schedule.append(sched + 1)
            circuit.append("TICK", [], [])

        # Add measurement operations
        self._add_extended_basis_operation(circuit, "M", meas_timestep_operations)
        schedule.append(MEASUREMENT_SCHEDULE)

        # Filter out unused qubits
        kept_data_qubits = [qubits.data_qubits[i] for i in used_data_qubit_indices]
        new_plaquette_qubits = PlaquetteQubits(kept_data_qubits, qubits.syndrome_qubits)
        unfiltered_circuit = ScheduledCircuit.from_circuit(circuit, schedule, qubits.qubit_map)
        filtered_circuit = unfiltered_circuit.filter_by_qubits(new_plaquette_qubits.all_qubits)

        # Return the plaquette
        return Plaquette(
            name=str(rpng_description),
            qubits=new_plaquette_qubits,
            circuit=filtered_circuit,
            mergeable_instructions=(
                RESET_INSTRUCTION_NAMES | MEASUREMENT_INSTRUCTION_NAMES | {"H"}
            ),
            debug_information=PlaquetteDebugInformation(rpng_description),
        )



from __future__ import annotations

import argparse
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Final

import stim
from typing_extensions import override

from tqec._cli.subcommands.base import TQECSubCommand
from tqec.utils.exceptions import TQECError


def has_program(name: str) -> bool:
    """Check if the provided ``name`` corresponds to a usable executable on the host system."""
    return shutil.which(name) is not None


def generate_animation(
    out_file: Path,
    framerate: int,
    image_directory: Path,
    image_glob_pattern: str = "*.svg",
    overwrite: bool = True,
) -> None:
    """Generate an animation from a directory containing images.

    Args:
        out_file: filepath to save the animation to.
        framerate: how many images per seconds should be displayed in the animated video.
        image_directory: a directory containing images following the provided
            ``image_glob_pattern``. Images should be named appropriately to be sorted.
        image_glob_pattern: file pattern used to get the list of all the images that should be
            in the saved video.
        overwrite: if ``True`` and ``out_file`` already exists, it will be overwritten. Else, the
            file will not be updated.

    Raises:
        TQECError: if ``ffmpeg`` is not available in the host system.

    """
    if not has_program("ffmpeg"):
        raise TQECError(
            "ffmpeg is needed to generate an animation, but could not find it. "
            "Make sure you have ffmpeg installed and that it is accessible."
        )
    overwrite_str = "-y" if overwrite else "-n"
    command = (
        f"ffmpeg {out_file} {overwrite_str} -framerate {framerate} -pattern_type "
        f"glob -i {image_directory}/{image_glob_pattern} -vf scale=1024:-1 -c:v "
        "libx264 -vf format=yuv420p -vf pad=ceil(iw/2)*2:ceil(ih/2)*2 -filter_complex "
        "[0]split=2[bg][fg];[bg]drawbox=c=white@1:t=fill[bg];[bg][fg]overlay=format=auto"
    )
    result = subprocess.run(command.split(), capture_output=True, check=False)
    # Print the path of the generated video on success.
    if result.returncode == 0:
        print(f"Video successfully generated at '{out_file}'.")
    else:
        linesep = "=" * 40
        print(f"Error when generating the video. Returned {result.returncode}.")
        print("Full stdout:")
        print(linesep)
        print(result.stdout.decode("utf-8"))
        print(linesep)
        print("Full stderr:")
        print(linesep)
        print(result.stderr.decode("utf-8"))
        print(linesep)


class VisualisationTQECSubCommand(TQECSubCommand):
    _STYLE_TO_STIM_DIAGRAM: Final[dict[str, str]] = {
        "timeslice": "timeslice-svg",
        "detslice": "detslice-svg",
        "detslice-with-ops": "detslice-with-ops-svg",
    }

    @staticmethod
    @override
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        parser: argparse.ArgumentParser = main_parser.add_parser(
            "viz", description="Generating images and videos from quantum circuits."
        )
        parser.add_argument(
            "-i",
            "--in_file",
            type=Path,
            required=True,
            help="Quantum circuit in Stim format that should be visualised.",
        )
        parser.add_argument(
            "-o",
            "--out_file",
            type=Path,
            required=True,
            help="Output filename that will be used to write the visualisation.",
        )
        parser.add_argument(
            "-t",
            "--ticks",
            default="ALL",
            help=(
                "A colon separated pair a:b where a < b representing the TICKS "
                "to include in the visualisation. Defaults to ALL, meaning all "
                "TICKS."
            ),
        )
        parser.add_argument(
            "--with_noise",
            action="store_true",
            help="If provided, noise instructions are included in the visualisation.",
        )
        parser.add_argument(
            "--style",
            choices=VisualisationTQECSubCommand._STYLE_TO_STIM_DIAGRAM.keys(),
            default="circ",
            help="Type of diagram to generate for the visualisation.",
        )
        parser.add_argument(
            "--anim",
            action="store_true",
            help=(
                "If provided, an animation (video) is generated. Else, an "
                "image containing the needed TICKs is generated."
            ),
        )
        parser.add_argument(
            "-f",
            "--framerate",
            type=int,
            default=5,
            help=(
                "Number of TICKS per seconds in the returned animation. Only "
                "has effect when '--anim' is provided."
            ),
        )
        parser.set_defaults(func=VisualisationTQECSubCommand.execute)

    @staticmethod
    @override
    def execute(args: argparse.Namespace) -> None:
        # Getting the quantum circuit
        circuit = stim.Circuit.from_file(args.in_file)
        if not args.with_noise:
            circuit = circuit.without_noise()
        # Computing the TICKs
        start_tick, end_tick = 0, circuit.num_ticks + 1
        if args.ticks != "ALL":
            start_tick, end_tick = tuple(int(s) for s in args.ticks.split(":"))
        # Getting the diagram style
        style = VisualisationTQECSubCommand._STYLE_TO_STIM_DIAGRAM[args.style]

        if args.anim:
            with tempfile.TemporaryDirectory() as tmpdirname:
                # Step one, generate the individual TICKs
                print("Generating individual images...")
                for t in range(start_tick, end_tick):
                    with open(f"{tmpdirname}/{t:0>4}.svg", "w") as f:
                        f.write(str(circuit.diagram(style, tick=t)))
                # Step two, generate a video with ffmpeg
                print("Generating animation (this may take some time)...")
                generate_animation(args.out_file, args.framerate, Path(tmpdirname))
        else:
            with open(args.out_file, "w") as f:
                f.write(str(circuit.diagram(style, tick=range(start_tick, end_tick))))



from collections.abc import Callable, Iterable, Iterator
from pathlib import Path

import sinter
import stim

from tqec.compile.detectors.database import DetectorDatabase
from tqec.compile.graph import TopologicalComputationGraph
from tqec.utils.noise_model import NoiseModel
from tqec.utils.paths import DEFAULT_DETECTOR_DATABASE_PATH


def generate_stim_circuits_with_detectors(
    compiled_graph: TopologicalComputationGraph,
    ks: Iterable[int],
    ps: Iterable[float],
    noise_model_factory: Callable[[float], NoiseModel],
    manhattan_radius: int,
    detector_database: DetectorDatabase | None = None,
    database_path: str | Path = DEFAULT_DETECTOR_DATABASE_PATH,
    do_not_use_database: bool = False,
    only_use_database: bool = False,
) -> Iterator[tuple[stim.Circuit, int, float]]:
    """Generate stim circuits in parallel.

    This function generate the `stim.Circuit` instances for all the combinations
    of the given `ks` and `ps` with a noise model that depends on `p` and
    computed with `noise_model_factory`.

    It is equivalent to:

    .. code-block:: python

        for p in ps:
            for k in ks:
                yield (
                    compiled_graph.generate_stim_circuit(k, noise_model_factory(p)), k, p
                )

    except that the order in which the results are returned is not guaranteed.

    Args:
        compiled_graph: computation to export to `stim.Circuit` instances.
        ks: values of `k` to consider.
        ps: values of `p`, the noise strength, to consider.
        noise_model_factory: a callable that builds a noise model from an input
            strength `p`.
        manhattan_radius: radius used to automatically compute detectors. The
            best value to set this argument to is the minimum integer such that
            flows generated from any given reset/measurement, from any plaquette
            at any spatial/temporal place in the QEC computation, do not
            propagate outside of the qubits used by plaquettes spatially located
            at maximum `manhattan_radius` plaquettes from the plaquette the
            reset/measurement belongs to (w.r.t. the Manhattan distance).
            Default to 2, which is sufficient for regular surface code. If
            negative, detectors are not computed automatically and are not added
            to the generated circuits.
        detector_database: an instance to retrieve from / store in detectors
            that are computed as part of the circuit generation. If not given,
            the detectors are retrieved from/stored in the provided
            ``database_path``.
        database_path: specify where to save to after the calculation.
            This defaults to :class:`.DEFAULT_DETECTOR_DATABASE_PATH` if
            not specified. If ``detector_database`` is not passed in, the code attempts to
            retrieve the database from this location. The user may pass in the path
            either in str format, or as a Path instance.
        do_not_use_database: if ``True``, even the default database will not be used.
        only_use_database: if ``True``, only detectors from the database
            will be used. An error will be raised if a situation that is not
            registered in the database is encountered.

    Yields:
        a tuple containing the resulting circuit, the value of `k` that
        corresponds to the returned circuit and the value of `p` that corresponds
        to the returned circuit.

    """
    noise_models = {p: noise_model_factory(p) for p in ps}
    circuits = {
        k: compiled_graph.generate_stim_circuit(
            k,
            manhattan_radius=manhattan_radius,
            detector_database=detector_database,
            database_path=database_path,
            do_not_use_database=do_not_use_database,
            only_use_database=only_use_database,
        )
        for k in ks
    }
    yield from (
        (nm.noisy_circuit(circuit), k, p)
        for k, circuit in circuits.items()
        for p, nm in noise_models.items()
    )


def generate_sinter_tasks(
    compiled_graph: TopologicalComputationGraph,
    ks: Iterable[int],
    ps: Iterable[float],
    noise_model_factory: Callable[[float], NoiseModel],
    manhattan_radius: int,
    detector_database: DetectorDatabase | None = None,
    database_path: str | Path = DEFAULT_DETECTOR_DATABASE_PATH,
    do_not_use_database: bool = False,
    only_use_database: bool = False,
) -> Iterator[sinter.Task]:
    """Generate `sinter.Task` instances from the provided parameters.

    This function generate the `sinter.Task` instances for all the combinations
    of the given `ks` and `ps` with a noise model that depends on `p` and
    computed with `noise_model_factory`.

    Args:
        compiled_graph: computation to export to `stim.Circuit` instances.
        ks: values of `k` to consider.
        ps: values of `p`, the noise strength, to consider.
        noise_model_factory: a callable that builds a noise model from an input
            strength `p`.
        manhattan_radius: radius used to automatically compute detectors. The
            best value to set this argument to is the minimum integer such that
            flows generated from any given reset/measurement, from any plaquette
            at any spatial/temporal place in the QEC computation, do not
            propagate outside of the qubits used by plaquettes spatially located
            at maximum `manhattan_radius` plaquettes from the plaquette the
            reset/measurement belongs to (w.r.t. the Manhattan distance).
            Default to 2, which is sufficient for regular surface code. If
            negative, detectors are not computed automatically and are not added
            to the generated circuits.
        detector_database: an instance to retrieve from / store in detectors
            that are computed as part of the circuit generation. If not given,
            the detectors are retrieved from/stored in the provided
            ``database_path``.
        database_path: specify where to save to after the calculation.
            This defaults to :data:`.DEFAULT_DETECTOR_DATABASE_PATH` if
            not specified. If ``detector_database`` is not passed in, the code attempts to
            retrieve the database from this location. The user may pass in the path
            either in str format, or as a Path instance.
        do_not_use_database: if ``True``, even the default database will not be used.
        only_use_database: if ``True``, only detectors from the database
            will be used. An error will be raised if a situation that is not
            registered in the database is encountered.

    Yields:
        tasks to be collected by a call to `sinter.collect`.

    """
    yield from (
        sinter.Task(
            circuit=circuit,
            json_metadata={"d": 2 * k + 1, "r": 2 * k + 1, "p": p},
        )
        for circuit, k, p in generate_stim_circuits_with_detectors(
            compiled_graph,
            ks,
            ps,
            noise_model_factory,
            manhattan_radius,
            detector_database,
            database_path=database_path,
            do_not_use_database=do_not_use_database,
            only_use_database=only_use_database,
        )
    )



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, cast

from tqec.circuit.schedule.schedule import Schedule
from tqec.utils.enums import Basis


class PauliBasis(Enum):
    X = "x"
    Y = "y"
    Z = "z"

    def __str__(self) -> str:
        return self.value  # pragma: no cover

    def to_extended_basis(self) -> ExtendedBasis:
        """Return ``self`` as an extended basis."""
        return ExtendedBasis(self.value)


class ExtendedBasis(Enum):
    X = PauliBasis.X.value
    Y = PauliBasis.Y.value
    Z = PauliBasis.Z.value
    H = "h"

    def __str__(self) -> str:
        return self.value  # pragma: no cover


@dataclass(frozen=True)
class RPNG:
    """Represent a single ``RPNG`` string.

    ## Format specification

    The ``RPNG`` string is a standard format used in ``tqec`` to unambiguously
    describe the action(s) being performed on a single data qubit. It is a
    4-character string. See each attribute docstring for more details on the
    possible values for each character.

    ## Example

    The following ::

        -z1-
        rpng

    represents a data-qubit with a ``CZ`` gate applied at timestep ``1``.

    Attributes:
        r: reset basis (``x``, ``y`` or ``z``), ``h`` or ``-``.
        p: controlled operation target basis (``x`` means ``CNOT`` controlled
            on the ancilla and targeting the data qubit, ``y`` means ``CY``,
            ``z`` means ``CZ``).
        n: time step at which the 2-qubit operation described by ``p`` should
            be applied. Should be a 1-digit positive integer, typically in
            ``[1, 5]``.
        g: measure basis (``x``, ``y`` or ``z``), ``h`` or ``-``.

    """

    r: ExtendedBasis | None
    p: PauliBasis | None
    n: int | None
    g: ExtendedBasis | None

    @classmethod
    def from_string(cls, rpng_string: str) -> RPNG:
        """Initialize the RPNG object from a 4-character string.

        Raises:
            ValueError: if an invalid ``rpng_string`` is provided.

        """
        if len(rpng_string) != 4:
            raise ValueError("The rpng string must be exactly 4-character long.")
        r_str, p_str, n_str, g_str = tuple(rpng_string)
        # Convert the characters into the enum attributes (or raise error).
        r = ExtendedBasis(r_str) if r_str in ExtendedBasis._value2member_map_ else None
        p = PauliBasis(p_str) if p_str in PauliBasis._value2member_map_ else None
        n = int(n_str) if n_str.isdigit() else None
        g = ExtendedBasis(g_str) if g_str in ExtendedBasis._value2member_map_ else None
        # Raise error if anythiong but '-' was used to indicate None.
        if not r and r_str != "-":
            raise ValueError("Unacceptable character for the R field.")
        if not p and p_str != "-":
            raise ValueError("Unacceptable character for the P field.")
        if not n and n_str != "-":
            raise ValueError("Unacceptable character for the N field.")
        if not g and g_str != "-":
            raise ValueError("Unacceptable character for the G field.")
        return cls(r, p, n, g)

    def get_r_op(self) -> str | None:
        """Get the reset operation or Hadamard."""
        op = self.r
        if op is None:
            return None
        elif op.value in PauliBasis._value2member_map_:
            return f"R{op.value.upper()}"
        else:
            return f"{op.value.upper()}"

    def get_g_op(self) -> str | None:
        """Get the measurement operation or Hadamard."""
        op = self.g
        if op is None:
            return None
        elif op.value in PauliBasis._value2member_map_:
            return f"M{op.value.upper()}"
        else:
            return f"{op.value.upper()}"

    @property
    def is_null(self) -> bool:
        """Check if the RPNG object is null, i.e. all fields are None."""
        return str(self) == "----"

    def __str__(self) -> str:
        r = "-" if self.r is None else self.r.value
        p = "-" if self.p is None else self.p.value
        n = "-" if self.n is None else self.n
        g = "-" if self.g is None else self.g.value
        return f"{r}{p}{n}{g}"


@dataclass(frozen=True)
class RG:
    """Reduced format to represent syndrome qubit operations.

    The ``RG`` format is simply the ``RPNG`` format where ``P`` is
    unconditionally ``-`` and ``N`` is unset.

    Attributes:
        r: reset basis (``x``, ``y`` or ``z``), ``h`` or ``-``.
        g: measure basis (``x``, ``y`` or ``z``), ``h`` or ``-``.

    """

    r: PauliBasis | None
    g: PauliBasis | None

    @classmethod
    def from_string(cls, rg_string: str) -> RG:
        """Initialize the ``RG`` object from a 2-character string."""
        if len(rg_string) != 2:
            raise ValueError("The RG string must be exactly 2-character long.")
        r_str, g_str = tuple(rg_string)

        try:
            r = None if r_str == "-" else PauliBasis(r_str)
            g = None if g_str == "-" else PauliBasis(g_str)
            return cls(r, g)
        except ValueError as err:
            raise ValueError(f"Invalid RG string: '{rg_string}'.") from err

    def __str__(self) -> str:
        return f"{'-' if self.r is None else self.r.value}{'-' if self.g is None else self.g.value}"


@dataclass(frozen=True)
class RPNGDescription:
    """Organize the description of a plaquette in RPNG format.

    The corners of the square plaquette are listed following the order:
    top-left, top-right, bottom-left, bottom-right.
    This forms a Z-shaped path on the plaquette corners:

    .. code-block:: text

        +------------+
        | 1 -----> 2 |
        |        /   |
        |      /     |
        |    ∟       |
        | 3 -----> 4 |
        +------------+

    If the ancilla RG description is not specified, it is assumed 'xx'

    Attributes:
        corners: one ``RPNG`` description for each of the four corners of the
            plaquette.
        ancilla: ``RG`` description of the syndrome qubit.

    """

    corners: tuple[RPNG, RPNG, RPNG, RPNG]
    ancilla: RG = field(default=RG(PauliBasis.X, PauliBasis.X))

    def __post_init__(self) -> None:
        """Validate the initialization arguments.

        Constraints:
        - the n values for the corners must be unique
        - the n values for the corners must be larger than 0

        """
        times = []
        for rpng in self.corners:
            if rpng.n:
                if rpng.n < 1:
                    raise ValueError("All n values must be larger than 0.")
                times.append(rpng.n)
        if len(times) != len(set(times)):
            raise ValueError("The n values for the corners must be unique.")

    @classmethod
    def from_string(cls, corners_rpng_string: str) -> RPNGDescription:
        """Initialize the RPNGDescription object from a (16+3)-character string."""
        rpng_objs = tuple([RPNG.from_string(s) for s in corners_rpng_string.split(" ")])
        if len(rpng_objs) != 4:
            raise ValueError("There must be 4 corners in the RPNG description.")
        return cls(cast(tuple[RPNG, RPNG, RPNG, RPNG], rpng_objs))

    @classmethod
    def from_extended_string(cls, ancilla_and_corners_rpng_string: str) -> RPNGDescription:
        """Initialize the RPNGDescription object from a (16+3)-character string."""
        values = ancilla_and_corners_rpng_string.split(" ")
        ancilla_rg = RG.from_string(values[0])
        rpng_objs = tuple([RPNG.from_string(s) for s in values[1:]])
        if len(rpng_objs) != 4:
            raise ValueError("There must be 4 corners in the RPNG description.")
        return cls(cast(tuple[RPNG, RPNG, RPNG, RPNG], rpng_objs), ancilla_rg)

    @classmethod
    def from_basis_and_schedule(
        cls,
        basis: Basis,
        schedule: Sequence[int] | Schedule,
        reset: PauliBasis | None = None,
        measurement: ExtendedBasis | None = None,
    ) -> RPNGDescription:
        """Initialize the RPNGDescription object from a basis and a schedule."""
        r = "-" if reset is None else reset.value
        m = "-" if measurement is None else measurement.value
        rpng_objs = tuple([RPNG.from_string(f"{r}{basis.value.lower()}{s}{m}") for s in schedule])
        if len(rpng_objs) != 4:
            raise ValueError("There must be 4 corners in the RPNG description.")
        return cls(cast(tuple[RPNG, RPNG, RPNG, RPNG], rpng_objs))

    @staticmethod
    def empty() -> RPNGDescription:
        """Return a description of the empty plaquette."""
        return RPNGDescription.from_extended_string("-- ---- ---- ---- ----")

    def get_r_op(self, data_idx: int) -> str | None:
        """Get the reset operation or Hadamard for the specific data qubit."""
        return self.corners[data_idx].get_r_op()

    def get_n(self, data_idx: int) -> int | None:
        """Get the time of the 2Q gate involving the specific data qubit."""
        return self.corners[data_idx].n

    def get_g_op(self, data_idx: int) -> str | None:
        """Get the measurement operation or Hadamard for the specific data qubit."""
        return self.corners[data_idx].get_g_op()

    @property
    def has_reset(self) -> bool:
        """Return ``True`` if ``self`` contains at least one corner with a reset."""
        return any(
            corner.get_r_op()
            not in {
                None,
                "H",
            }
            for corner in self.corners
        )

    @property
    def has_measurement(self) -> bool:
        """Return ``True`` if ``self`` contains at least one corner with a measurement."""
        return any(
            corner.get_g_op()
            not in {
                None,
                "H",
            }
            for corner in self.corners
        )

    def __str__(self) -> str:
        return " ".join(str(rpng) for rpng in self.corners)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the RPNG description.

        The dictionary is intended to be used as a JSON object.

        """
        return {
            "corners": [str(rpng) for rpng in self.corners],
            "ancilla": str(self.ancilla),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> RPNGDescription:
        """Return a RPNGDescription object from its dictionary representation.

        Args:
            data: dictionary with the keys ``corners`` and ``ancilla``.

        Returns:
            a new instance of :class:`RPNGDescription` with the provided
            ``corners`` and ``ancilla``.

        """
        assert len(data["corners"]) == 4, "There must be 4 corners in the RPNG description."
        corners = data["corners"]
        corners = (
            RPNG.from_string(corners[0]),
            RPNG.from_string(corners[1]),
            RPNG.from_string(corners[2]),
            RPNG.from_string(corners[3]),
        )
        ancilla = RG.from_string(data["ancilla"])
        return RPNGDescription(corners, ancilla)



from dataclasses import dataclass

from tqec.interop.color import TQECColor


@dataclass(frozen=True)
class DrawerConfiguration:
    """Drawing configuration with sensible defaults that can be modified by the user.

    Attributes:
        stroke_width: default stroke width used for all the drawing, except for
            error crosses.
        stroke_color: default stroke color used for all the drawing, except for
            error crosses.
        thin_stroke_color: color used to draw stroke that should appear thinner,
            for example between the UP and DOWN plaquette of an extended plaquette.
        font_size: default font size of all the text in the drawing except the
            overlaid "Moments: XX -> YY" text.
        hook_error_line_lerp_coefficient: a coefficient in [0, 1] that defines how close
            the hook error line should be from the borders of the plaquette. As an
            implementation detail, that number is also the hook error line length
            compared to the full plaquette width or height. A value of ``0.9`` means
            that the hook error line is closer to the plaquette border than to the
            plaquette center.
        plaquette_overflow_lerp_coefficient: coefficient in [0, 1] that defines how
            some plaquettes overflow from the regular convex hull of the involved
            data qubits. This is currently used for triangular plaquettes, but might
            be used for other plaquettes later. A value of ``0`` means "no overflow"
            whereas a value of ``1`` means "full overflow up to the missing data qubit".
        text_lerp_coefficient: coefficient in [0, 1] that defines where data-qubit
            interaction orders are written in the plaquette. A value of ``0`` means
            "on their respective data-qubit" whereas a value of ``1`` means "on the
            plaquette center".
        mixed_basis_color: color used to plot plaquettes measuring a non-uniform
            Pauli basis (e.g. ``ZXXZ``).
        observable_fill_opacity: opacity of the color used to fill observable stars.
        observable_stroke_width_multiplier: multiplier used on the plaquette width to obtain the
            stroke width that should be used to draw observables.
        observable_star_color: color used to draw the observable stars.
        reset_square_radius: representative size (in SVG coordinates) of the square used
            to represent data-qubit reset.
        measurement_circle_radius: representative size (in SVG coordinates) of the circle
            used to represent data-qubit measurement.

    """

    # Default stroke parameters
    stroke_width: float = 0.01
    stroke_color: str = "black"
    thin_stroke_color: str = "grey"
    # Default font parameters
    font_size: float = 0.1
    # Default placement parameters, used as coefficients to perform a linear interpolation
    hook_error_line_lerp_coefficient: float = 0.9
    plaquette_overflow_lerp_coefficient: float = 0.2
    text_lerp_coefficient: float = 0.8
    # Default colors used for different contents when drawing
    mixed_basis_color: str = "gray"
    # Observable-related drawing parameters
    observable_fill_opacity: float = 0.5
    observable_stroke_width_multiplier: float = 0.01
    observable_star_color: TQECColor = TQECColor.H
    # Default size of some of the elements drawn
    reset_square_radius: float = 0.05
    measurement_circle_radius: float = 0.1



import argparse
import sys

from tqec._cli.subcommands.check_dae import CheckDaeTQECSubCommand
from tqec._cli.subcommands.dae2circuits import Dae2CircuitsTQECSubCommand
from tqec._cli.subcommands.dae2observables import Dae2ObservablesTQECSubCommand
from tqec._cli.subcommands.run_example import RunExampleTQECSubCommand
from tqec._cli.subcommands.viz import VisualisationTQECSubCommand


def main() -> None:
    """Entry point of TQEC CLI."""
    parser = argparse.ArgumentParser(
        prog="tqec",
        description="The main tqec command-line tool.",
    )
    subparser = parser.add_subparsers(title="subcommands")

    Dae2ObservablesTQECSubCommand.add_subcommand(subparser)
    CheckDaeTQECSubCommand.add_subcommand(subparser)
    Dae2CircuitsTQECSubCommand.add_subcommand(subparser)
    RunExampleTQECSubCommand.add_subcommand(subparser)
    VisualisationTQECSubCommand.add_subcommand(subparser)

    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])
    args.func(args)



"""Defines all the necessary data-structures to represent a plaquette.

This package defines one of the core class of the `tqec` library:
:class:`~.plaquette.Plaquette`.
The :class:`~.plaquette.Plaquette` class represents what is commonly called a
"plaquette" in quantum error correction and is basically a
:class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` instance
representing the computation defining the plaquette.

Because we do not have a module to perform simple geometry operations on qubits
(yet), the :mod:`tqec.plaquette.qubit` module is providing classes to represent
the qubits a plaquette is applied to and perform some operations on them (e.g.,
get the qubits on a specific side of the plaquette).

"""

from ..utils.frozendefaultdict import FrozenDefaultDict as FrozenDefaultDict
from .enums import PlaquetteOrientation as PlaquetteOrientation
from .enums import PlaquetteSide as PlaquetteSide
from .plaquette import Plaquette as Plaquette
from .plaquette import Plaquettes as Plaquettes
from .qubit import PlaquetteQubits as PlaquetteQubits
from .qubit import SquarePlaquetteQubits as SquarePlaquetteQubits
from .rpng import RG as RG
from .rpng import RPNG as RPNG
from .rpng import RPNGDescription as RPNGDescription



from typing import Final

EXTENDED_PLAQUETTE_SCHEDULES: Final[dict[bool, tuple[int, int, int, int]]] = {
    False: (2, 4, 3, 5),
    True: (5, 3, 4, 2),
}
VERTICAL_HOOK_SCHEDULES: Final[dict[bool, tuple[int, int, int, int]]] = {
    False: (1, 4, 3, 5),
    True: (5, 3, 4, 1),
}
HORIZONTAL_HOOK_SCHEDULES: Final[dict[bool, tuple[int, int, int, int]]] = {
    False: (1, 2, 3, 5),
    True: (5, 3, 2, 1),
}
